from typing import List, Dict, Tuple, Union

from aioinstagrapi.extractors import (
    extract_hashtag_v1,
    extract_location,
    extract_track,
    extract_user_short,
)
from aioinstagrapi.models import Hashtag, Location, Track, UserShort
from aioinstagrapi.utils import generate_uuid

class FbSearchMixin:
    async def fbsearch_places(
        self, query: str, lat: float = 40.74, lng: float = -73.94
    ) -> List[Location]:
        params = {
            "search_surface": "places_search_page",
            "timezone_offset": self.timezone_offset,
            "lat": lat,
            "lng": lng,
            "count": 30,
            "query": query,
        }
        result = await self.private_request("fbsearch/places/", params=params)
        locations = []
        for item in result["items"]:
            locations.append(extract_location(item["location"]))
        return locations

    async def fbsearch_topsearch_flat(self, query: str) -> List[dict]:
        params = {
            "search_surface": "top_search_page",
            "context": "blended",
            "timezone_offset": self.timezone_offset,
            "count": 30,
            "query": query,
        }
        result = await self.private_request("fbsearch/topsearch_flat/", params=params)
        return result["list"]

    async def search_users(self, query: str) -> List[UserShort]:
        params = {
            "search_surface": "user_search_page",
            "timezone_offset": self.timezone_offset,
            "count": 30,
            "q": query,
        }
        result = await self.private_request("users/search/", params=params)
        return [extract_user_short(item) for item in result["users"]]

    async def search_music(self, query: str) -> List[Track]:
        params = {
            "query": query,
            "browse_session_id": generate_uuid(),
        }
        result = await self.private_request("music/audio_global_search/", params=params)
        return [extract_track(item["track"]) for item in result["items"]]

    async def search_hashtags(self, query: str) -> List[Hashtag]:
        params = {
            "search_surface": "hashtag_search_page",
            "timezone_offset": self.timezone_offset,
            "count": 30,
            "q": query,
        }
        result = await self.private_request("tags/search/", params=params)
        return [extract_hashtag_v1(ht) for ht in result["results"]]

    async def fbsearch_suggested_profiles(self, user_id: str) -> List[UserShort]:
        params = {
            "target_user_id": user_id,
            "include_friendship_status": "true",
        }
        result = await self.private_request("fbsearch/accounts_recs/", params=params)
        return result["users"]

    async def fbsearch_recent(self) -> List[Tuple[int, Union[UserShort, Hashtag, Dict]]]:
        """
        Retrieves recently searched results

        Returns
        -------
        List[Tuple[int, Union[UserShort, Hashtag, Dict]]]
            Returns list of Tuples where first value is timestamp of searh, second is retrived result
        """
        result = await self.private_request("fbsearch/recent_searches/")
        assert result.get("status", "") == "ok", "Failed to retrieve recent searches"

        data = []
        for item in result.get("recent", []):
            if "user" in item.keys():
                data.append(
                    (item.get("client_time", None), extract_user_short(item["user"]))
                )
            if "hashtag" in item.keys():
                hashtag = item.get("hashtag")
                hashtag["media_count"] = hashtag.pop("formatted_media_count")
                data.append((item.get("client_time", None), Hashtag(**hashtag)))
            if "keyword" in item.keys():
                data.append((item.get("client_time", None), item["keyword"]))
        return data



from typing import Any, Dict
import httpx
from pathlib import Path
from urllib.parse import urlparse

import aiofiles

from aioinstagrapi.extractors import extract_track
from aioinstagrapi.models import Track
from aioinstagrapi.utils import json_value
from aioinstagrapi.exceptions import ClientError, TrackNotFound


class TrackMixin:
    async def track_download_by_url(
        self, url: str, filename: str = "", folder: Path = ""
    ) -> Path:
        """
        Download track by URL

        Parameters
        ----------
        url: str
            URL for a track
        filename: str, optional
            Filename for the track
        folder: Path, optional
            Directory in which you want to download the track, default is "" and will download the files to working directory

        Returns
        -------
        Path
            Path for the file downloaded
        """
        fname = urlparse(url).path.rsplit("/", 1)[1].strip()
        assert fname, """The URL must contain the path to the file (m4a or mp3)."""
        filename = "%s.%s" % (filename, fname.rsplit(".", 1)[1]) if filename else fname
        path = Path(folder) / filename
        async with httpx.AsyncClient(timeout=self.request_timeout) as client:
            async with client.stream("GET", url) as response:
                response.raise_for_status()
                async with aiofiles.open(path, "wb") as f:
                    async for chunk in response.aiter_bytes():
                        f.write(chunk)
        return path.resolve()

    async def _track_request(self, data: Dict[str, Any]) -> Dict:
        try:
            result = await self.private_request("clips/music/", data)
        except ClientError as e:
            if not self.last_json:
                kw = {
                    k: v
                    for k, v in data.items()
                    if k in {"music_canonical_id", "original_sound_audio_asset_id"}
                }
                raise TrackNotFound(**kw)
            raise e
        return result

    async def track_info_by_canonical_id(self, music_canonical_id: str) -> Track:
        """
        Get Track by music_canonical_id

        Parameters
        ----------
        music_canonical_id: str
            Unique identifier of the track

        Returns
        -------
        Track
            An object of Track type
        """
        data = {
            "tab_type": "clips",
            "referrer_media_id": "",
            "_uuid": self.uuid,
            "music_canonical_id": str(music_canonical_id),
        }
        result = await self.private_request("clips/music/", data)
        track = json_value(result, "metadata", "music_info", "music_asset_info")
        return extract_track(track)

    async def track_info_by_id(self, track_id: str, max_id: str = "") -> Dict:
        """
        Get Track by id

        Parameters
        ----------
        track_id: str
            Unique identifier of the track

        Returns
        -------
        Dict
            Raw insta response json
        """
        data = {
            "audio_cluster_id": track_id,
            "original_sound_audio_asset_id": track_id,
        }
        if max_id:
            data["max_id"] = max_id
        return await self._track_request(data)



import json
import random
import tempfile
import time
import asyncio
import aiofiles
from pathlib import Path
from typing import Dict, List
from uuid import uuid4
from aioinstagrapi import config
from aioinstagrapi.exceptions import ClientError, ClipConfigureError, ClipNotUpload
from aioinstagrapi.extractors import extract_media_v1
from aioinstagrapi.models import Location, Media, Track, Usertag
from aioinstagrapi.utils import date_time_original

try:
    from PIL import Image
except ImportError:
    raise Exception("You don't have PIL installed. Please install PIL or Pillow>=8.1.1")


class DownloadClipMixin:
    """
    Helpers to download CLIP videos
    """

    async def clip_download(self, media_pk: int, folder: Path = "") -> str:
        """
        Download CLIP video

        Parameters
        ----------
        media_pk: int
            PK for the album you want to download
        folder: Path, optional
            Directory in which you want to download the album,
            default is "" and will download the files to working
            directory.

        Returns
        -------
        str
        """
        return await self.video_download(media_pk, folder)

    async def clip_download_by_url(
        self, url: str, filename: str = "", folder: Path = ""
    ) -> str:
        """
        Download CLIP video using URL

        Parameters
        ----------
        url: str
            URL to download media from
        folder: Path, optional
            Directory in which you want to download the album,
            default is "" and will download the files to working
            directory.

        Returns
        -------
        str
        """
        return await self.video_download_by_url(url, filename, folder)


class UploadClipMixin:
    """
    Helpers to upload CLIP videos
    """

    async def clip_upload(
        self,
        path: Path,
        caption: str,
        thumbnail: Path = None,
        usertags: List[Usertag] = [],
        location: Location = None,
        configure_timeout: int = 10,
        feed_show: str = "1",
        extra_data: Dict[str, str] = {},
    ) -> Media:
        """
        Upload CLIP to Instagram

        Parameters
        ----------
        path: Path
            Path to CLIP file
        caption: str
            Media caption
        thumbnail: Path, optional
            Path to thumbnail for CLIP.
            Default value is None, and it generates a thumbnail
        usertags: List[Usertag], optional
            List of users to be tagged on this upload, default is empty list.
        location: Location, optional
            Location tag for this upload, default is none
        configure_timeout: int
            Timeout between attempt to configure media (set caption, etc), default is 10
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params,
            like {"share_to_facebook": 1}.

        Returns
        -------
        Media
            An object of Media class
        """
        path = Path(path)
        if thumbnail is not None:
            thumbnail = Path(thumbnail)
        upload_id = str(int(time.time() * 1000))
        thumbnail, width, height, duration = await analyze_video(path, thumbnail)
        waterfall_id = str(uuid4())
        # upload_name example: '1576102477530_0_7823256191'
        upload_name = "{upload_id}_0_{rand}".format(
            upload_id=upload_id, rand=random.randint(1000000000, 9999999999)
        )
        rupload_params = {
            "is_clips_video": "1",
            "retry_context": '{"num_reupload":0,"num_step_auto_retry":0,"num_step_manual_retry":0}',
            "media_type": "2",
            "xsharing_user_ids": json.dumps([self.user_id]),
            "upload_id": upload_id,
            "upload_media_duration_ms": str(int(duration * 1000)),
            "upload_media_width": str(width),
            "upload_media_height": str(height),
        }
        headers = {
            "Accept-Encoding": "gzip",
            "X-Instagram-Rupload-Params": json.dumps(rupload_params),
            "X_FB_VIDEO_WATERFALL_ID": waterfall_id,
            "X-Entity-Type": "video/mp4",
        }
        response = await self.private_request(
            domain="https://{domain}/rupload_igvideo/{name}".format(
                domain=config.API_DOMAIN, name=upload_name
            ),
            headers=headers,
        )
        self.request_log(response)
        if response.status_code != 200:
            raise ClipNotUpload(response=self.last_response, **self.last_json)
        async with aiofiles.open(path, "rb") as fp:
            clip_data = fp.read()
            clip_len = str(len(clip_data))
        headers = {
            "Offset": "0",
            "X-Entity-Name": upload_name,
            "X-Entity-Length": clip_len,
            "Content-Type": "application/octet-stream",
            "Content-Length": clip_len,
            **headers,
        }
        response = self.private_request(
            domain="https://{domain}/rupload_igvideo/{name}".format(
                domain=config.API_DOMAIN, name=upload_name
            ),
            data=clip_data,
            headers=headers,
        )
        self.request_log(response)
        if response.status_code != 200:
            raise ClipNotUpload(response=self.last_response, **self.last_json)
        # CONFIGURE
        # self.igtv_composer_session_id = self.generate_uuid()  #issue
        for attempt in range(50):
            self.logger.debug(f"Attempt #{attempt} to configure CLIP: {path}")
            await asyncio.sleep(configure_timeout)
            try:
                configured = await self.clip_configure(
                    upload_id,
                    thumbnail,
                    width,
                    height,
                    duration,
                    caption,
                    usertags,
                    location,
                    feed_show,
                    extra_data=extra_data,
                )
            except ClientError as e:
                if "Transcode not finished yet" in str(e):
                    """
                    Response 202 status:
                    {"message": "Transcode not finished yet.", "status": "fail"}
                    """
                    await asyncio.sleep(configure_timeout)
                    continue
                raise e
            else:
                if configured:
                    media = self.last_json.get("media")
                    await self.expose()
                    return extract_media_v1(media)
        raise ClipConfigureError(response=self.last_response, **self.last_json)

    async def clip_upload_as_reel_with_music(
        self,
        path: Path,
        caption: str,
        track: Track,
        extra_data: Dict[str, str] = {},
    ) -> Media:

        """
        Upload CLIP as reel with music metadata.
        It also add the music under the video, therefore a mute video is required.

        If you just want to add music metadata to your reel,
        just copy the extra data you find here and add it
        to the extra_data parameter of the clip_upload function.

        Parameters
        ----------
        path: Path
            Path to CLIP file
        caption: str
            Media caption
        track: Track
            The music track to be added to the video reel
            use cl.search_music(title)[0].dict()

        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Media
            A Media response from the call
        """
        tmpaudio = Path(tempfile.mktemp(".m4a"))
        tmpaudio = await self.track_download_by_url(track.uri, "track", tmpaudio.parent)
        try:
            highlight_start_time = track.highlight_start_times_in_ms[0]
        except IndexError:
            highlight_start_time = 0
        try:
            import moviepy.editor as mp
        except ImportError:
            raise Exception("Please install moviepy>=1.0.3 and retry")
        # get all media to create the reel
        video = mp.VideoFileClip(str(path))
        audio_clip = mp.AudioFileClip(str(tmpaudio))
        # set the start time of the audio and create the actual media
        start = highlight_start_time / 1000
        end = highlight_start_time / 1000 + video.duration
        audio_clip = audio_clip.subclip(start, end)
        video = video.set_audio(audio_clip)
        # save the media in tmp folder
        tmpvideo = Path(tempfile.mktemp(".mp4"))
        video.write_videofile(str(tmpvideo))
        # close the media
        try:
            video.close()
        except AttributeError:
            pass
        try:
            audio_clip.close()
        except AttributeError:
            pass
        # create the extra data to upload with it
        data = extra_data or {}
        data["clips_audio_metadata"] = (
            {
                "original": {"volume_level": 0.0},
                "song": {
                    "volume_level": 1.0,
                    "is_saved": "0",
                    "artist_name": track.display_artist,
                    "audio_asset_id": track.id,
                    "audio_cluster_id": track.audio_cluster_id,
                    "track_name": track.title,
                    "is_picked_precapture": "1",
                },
            },
        )
        data["music_params"] = {
            "audio_asset_id": track.id,
            "audio_cluster_id": track.audio_cluster_id,
            "audio_asset_start_time_in_ms": highlight_start_time,
            "derived_content_start_time_in_ms": 0,
            "overlap_duration_in_ms": 15000,
            "product": "story_camera_clips_v2",
            "song_name": track.title,
            "artist_name": track.display_artist,
            "alacorn_session_id": "null",
        }
        clip_upload = await self.clip_upload(tmpvideo, caption, extra_data=data)
        # remove the tmp files
        tmpvideo.unlink()
        tmpaudio.unlink()
        return clip_upload

    async def clip_configure(
        self,
        upload_id: str,
        thumbnail: Path,
        width: int,
        height: int,
        duration: int,
        caption: str,
        usertags: List[Usertag] = [],
        location: Location = None,
        feed_show: str = "1",
        extra_data: Dict[str, str] = {},
    ) -> Dict:
        """
        Post Configure CLIP (send caption, thumbnail and more to Instagram)

        Parameters
        ----------
        upload_id: str
            Unique identifier for a IGTV video
        thumbnail: Path
            Path to thumbnail for IGTV
        width: int
            Width of the video in pixels
        height: int
            Height of the video in pixels
        duration: int
            Duration of the video in seconds
        caption: str
            Media caption
        usertags: List[Usertag], optional
            List of users to be tagged on this upload, default is empty list.
        location: Location, optional
            Location tag for this upload, default is None
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        await self.photo_rupload(Path(thumbnail), upload_id)
        usertags = [
            {"user_id": tag.user.pk, "position": [tag.x, tag.y]} for tag in usertags
        ]
        data = {
            # "igtv_ads_toggled_on": "0",
            "filter_type": "0",
            "timezone_offset": str(self.timezone_offset),
            "media_folder": "ScreenRecorder",
            "location": self.location_build(location),
            "source_type": "4",
            # "title": title,
            "caption": caption,
            "usertags": json.dumps({"in": usertags}),
            "date_time_original": date_time_original(time.localtime()),
            "clips_share_preview_to_feed": feed_show,
            "upload_id": upload_id,
            # "igtv_composer_session_id": self.igtv_composer_session_id,
            "device": self.device,
            "length": duration,
            "clips": [{"length": duration, "source_type": "4"}],
            "extra": {"source_width": width, "source_height": height},
            "audio_muted": False,
            "poster_frame_index": 70,
            **extra_data,
        }
        return await self.private_request(
            "media/configure_to_clips/?video=1",
            self.with_default_data(data),
            with_signature=True,
        )


async def analyze_video(path: Path, thumbnail: Path = None) -> tuple:
    """
    Analyze and crop thumbnail if need

    Parameters
    ----------
    path: Path
        Path to the video
    thumbnail: Path
        Path to thumbnail for CLIP

    Returns
    -------
    Tuple
        A tuple with (thumbail path, width, height, duration)
    """
    try:
        import moviepy.editor as mp
    except ImportError:
        raise Exception("Please install moviepy>=1.0.3 and retry")

    print(f'Analyzing CLIP file "{path}"')
    video = mp.VideoFileClip(str(path))
    width, height = video.size
    if not thumbnail:
        thumbnail = f"{path}.jpg"
        print(f'Generating thumbnail "{thumbnail}"...')
        video.save_frame(thumbnail, t=(video.duration / 2))
        await crop_thumbnail(thumbnail)
    return thumbnail, width, height, video.duration


async def crop_thumbnail(path: Path) -> bool:
    """
    Analyze and crop thumbnail if need

    Parameters
    ----------
    path: Path
        Path to the video

    Returns
    -------
    bool
        A boolean value
    """
    im = Image.open(str(path))
    width, height = im.size
    offset = (height / 1.78) / 2
    center = width / 2
    # Crop the center of the image
    im = im.crop((center - offset, 0, center + offset, height))
    async with aiofiles.open(path, "w") as fp:
        im.save(fp)
        im.close()
    return True



import json
from typing import Dict
class PreLoginFlowMixin:
    """
    Helpers for pre login flow
    """

    async def pre_login_flow(self) -> bool:
        """
        Emulation mobile app behavior before login

        Returns
        -------
        bool
            A boolean value
        """
        # self.set_contact_point_prefill("prefill")
        # self.get_prefill_candidates(True)
        # self.set_contact_point_prefill("prefill")
        await self.get_tokens()
        await self.home_template()
        #await self.attestation()
        await self.mobile_config_login()
        # await self.sync_launcher(True)
        # self.sync_device_features(True)
        return True
    
    async def mobile_config_login(self):
        headers = {'X-Ig-Nav-Chain': 'com.bloks.www.caa.login.home_template:com.bloks.www.caa.login.home_template:1:warm_start:1703001760.393::,com.bloks.www.caa.login.save-credentials:com.bloks.www.caa.login.save-credentials:2:button:1703001792.449::',}
        data = {"bool_opt_policy":"0","mobileconfigsessionless":"","api_version":"3","unit_type":"1","query_hash":"4b4d5d0d4ef2f269eb5cd59adc5601462f8e01d5595b05e08913652a73d0604b","ts":"1703001789","device_id":self.uuid,"fetch_type":"ASYNC_FULL","family_device_id":self.phone_id}
        return await self.private_request(
            "launcher/mobileconfig/", data=data, headers=headers
        )

    async def home_template(self):
        data = {
            'params': json.dumps({"server_params":{"qe_device_id_server":self.uuid,"family_device_id_server":"","device_id_server":self.android_device_id}}),
            'bk_client_context': '{"bloks_version":"525157204187ff40ed117cb8039807a0b12137bc9d50211ed3b2e1d97be928a6","styles_id":"instagram"}',
            'bloks_versioning_id': '525157204187ff40ed117cb8039807a0b12137bc9d50211ed3b2e1d97be928a6',
        }
        return await self.private_request(domain='https://b.i.instagram.com/api/v1/bloks/apps/com.bloks.www.caa.login.home_template/', data=data, headers={"Host":"b.i.instagram.com"}, with_signature=False)

    async def get_tokens(self):
        data = {
        'normal_token_hash': '',
        'device_id': self.android_device_id,
        'custom_device_id': self.uuid,
        'fetch_reason': 'token_expired',
    }
        return await self.private_request(domain='https://b.i.instagram.com/api/v1/zr/dual_tokens/', data=data, headers={"Host":"b.i.instagram.com"}, with_signature=False)

    
    async def attestation(self):
        data = {
        'app_scoped_device_id': self.uuid,
        'key_hash': '',
            }
        return await self.private_request(domain='https://b.i.instagram.com/api/v1/attestation/create_android_keystore/', data=data, headers={"Host":"b.i.instagram.com"}, with_signature=False)
        
    
    async def get_prefill_candidates(self, login: bool = False) -> Dict:
        """
        Get prefill candidates value from Instagram

        Parameters
        ----------
        login: bool, optional
            Whether to login or not

        Returns
        -------
        bool
            A boolean value
        """
        data = {
            "android_device_id": self.android_device_id,
            "client_contact_points": '[{"type":"omnistring","value":"%s","source":"last_login_attempt"}]'
            % self.username,
            "phone_id": self.phone_id,
            "usages": '["account_recovery_omnibox"]',
            "logged_in_user_ids": "[]",  # "[\"123456789\",\"987654321\"]",
            "device_id": self.uuid,
        }
        # if login is False:
        data["_csrftoken"] = self.token
        return await self.private_request(
            "accounts/get_prefill_candidates/", data, login=login
        )

    async def sync_device_features(self, login: bool = False) -> Dict:
        """
        Sync device features to your Instagram account

        Parameters
        ----------
        login: bool, optional
            Whether to login or not

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        data = {
            "id": self.uuid,
            "server_config_retrieval": "1",
            # "experiments": config.LOGIN_EXPERIMENTS,
        }
        if login is False:
            data["_uuid"] = self.uuid
            data["_uid"] = self.user_id
            data["_csrftoken"] = self.token
        # headers={"X-DEVICE-ID": self.uuid}
        return await self.private_request("qe/sync/", data, login=login)

    async def sync_launcher(self, login: bool = False) -> Dict:
        """
        Sync Launcher

        Parameters
        ----------
        login: bool, optional
            Whether to login or not

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        data = {
            "id": self.uuid,
            "server_config_retrieval": "1",
        }
        if login is False:
            data["_uid"] = self.user_id
            data["_uuid"] = self.uuid
            data["_csrftoken"] = self.token
        return await self.private_request("launcher/sync/", data, login=login)

    async def set_contact_point_prefill(self, usage: str = "prefill") -> Dict:
        """
        Sync Launcher

        Parameters
        ----------
        usage: str, optional
            Default "prefill"

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        data = {
            "phone_id": self.phone_id,
            "usage": usage,
            # "_csrftoken": self.token
        }
        return await self.private_request("accounts/contact_point_prefill/", data, login=True)



from typing import List, Tuple

from aioinstagrapi.exceptions import CollectionNotFound
from aioinstagrapi.extractors import extract_collection, extract_media_v1
from aioinstagrapi.models import Collection, Media


class CollectionMixin:
    """
    Helpers for collection
    """

    async def collections(self) -> List[Collection]:
        """
        Get collections

        Returns
        -------
        List[Collection]
            A list of objects of Collection
        """
        next_max_id = ""
        total_items = []
        while True:
            try:
                result = await self.private_request(
                    "collections/list/",
                    params={
                        "collection_types": '["ALL_MEDIA_AUTO_COLLECTION","PRODUCT_AUTO_COLLECTION","MEDIA"]',
                        "max_id": next_max_id,
                    },
                )
            except Exception as e:
                self.logger.exception(e)
                return total_items
            for item in result["items"]:
                total_items.append(extract_collection(item))
            if not result.get("more_available"):
                return total_items
            next_max_id = result.get("next_max_id", "")
        return total_items

    async def collection_pk_by_name(self, name: str) -> int:
        """
        Get collection_pk by name

        Parameters
        ----------
        name: str
            Name of the collection

        Returns
        -------
        List[Collection]
            A list of objects of Collection
        """
        for item in self.collections():
            if item.name == name:
                return item.id
        raise CollectionNotFound(name=name)

    async def collection_medias_by_name(self, name: str) -> List[Collection]:
        """
        Get medias by collection name

        Parameters
        ----------
        name: str
            Name of the collection

        Returns
        -------
        List[Collection]
            A list of collections
        """

        return await self.collection_medias(await self.collection_pk_by_name(name))

    async def liked_medias(self, amount: int = 21, last_media_pk: int = 0) -> List[Media]:
        """
        Get media you have liked

        Parameters
        ----------
        amount: int, optional
            Maximum number of media to return, default is 21
        last_media_pk: int, optional
            Last PK user has seen, function will return medias after this pk. Default is 0
        Returns
        -------
        List[Media]
            A list of objects of Media
        """
        return await self.collection_medias("liked", amount, last_media_pk)

    async def collection_medias_v1_chunk(
        self, collection_pk: str, max_id: str = ""
    ) -> Tuple[List[Media], str]:
        """
        Get media in a collection by collection_pk

        Parameters
        ----------
        collection_pk: str
            Unique identifier of a Collection
        max_id: str, optional
            Cursor

        Returns
        -------
        Tuple[List[Media], str]
            A list of objects of Media and cursor
        """
        if isinstance(collection_pk, int) or collection_pk.isdigit():
            private_request_endpoint = f"feed/collection/{collection_pk}/"
        elif collection_pk.lower() == "liked":
            private_request_endpoint = "feed/liked/"
        else:
            private_request_endpoint = "feed/saved/posts/"

        params = {"include_igtv_preview": "false"}
        if max_id:
            params["max_id"] = max_id
        result = await self.private_request(private_request_endpoint, params=params)
        items = [extract_media_v1(m.get("media", m)) for m in result["items"]]
        return items, result.get("next_max_id", "")

    async def collection_medias_v1(
        self, collection_pk: str, amount: int = 21, last_media_pk: int = 0
    ) -> List[Media]:
        """
        Get media in a collection by collection_pk

        Parameters
        ----------
        collection_pk: str
            Unique identifier of a Collection
        amount: int, optional
            Maximum number of media to return, default is 21
        last_media_pk: int, optional
            Last PK user has seen, function will return medias after this pk. Default is 0

        Returns
        -------
        List[Media]
            A list of objects of Media
        """
        last_media_pk = last_media_pk and int(last_media_pk)
        total_items = []
        next_max_id = ""
        amount = int(amount)
        found_last_media_pk = False
        while True:
            items, next_max_id = await self.collection_medias_v1_chunk(
                collection_pk, max_id=next_max_id
            )
            for item in items:
                if last_media_pk and last_media_pk == item["media"]["pk"]:
                    found_last_media_pk = True
                    break
                total_items.append(item)
            if (amount and len(total_items) >= amount) or found_last_media_pk:
                break
            if not items or not next_max_id:
                break
        return total_items[:amount] if amount else total_items

    async def collection_medias(
        self, collection_pk: str, amount: int = 21, last_media_pk: int = 0
    ) -> List[Media]:
        """
        Get media in a collection by collection_pk

        Parameters
        ----------
        collection_pk: str
            Unique identifier of a Collection
        amount: int, optional
            Maximum number of media to return, default is 21
        last_media_pk: int, optional
            Last PK user has seen, function will return medias after this pk. Default is 0

        Returns
        -------
        List[Media]
            A list of objects of Media
        """
        return await self.collection_medias_v1(
            collection_pk, amount=amount, last_media_pk=last_media_pk
        )

    async def media_save(
        self, media_id: str, collection_pk: int = None, revert: bool = False
    ) -> bool:
        """
        Save a media to collection

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        collection_pk: int
            Unique identifier of a Collection
        revert: bool, optional
            If True then save to collection, otherwise unsave

        Returns
        -------
        bool
            A boolean value
        """
        assert self.user_id, "Login required"
        media_id = await self.media_pk(media_id)
        data = {
            "module_name": "feed_timeline",
            "radio_type": "wifi-none",
        }
        if collection_pk:
            data["added_collection_ids"] = f"[{int(collection_pk)}]"
        name = "unsave" if revert else "save"
        result = await self.private_request(
            f"media/{media_id}/{name}/", self.with_action_data(data)
        )
        return result["status"] == "ok"

    async def media_unsave(self, media_id: str, collection_pk: int = None) -> bool:
        """
        Unsave a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        collection_pk: int
            Unique identifier of a Collection

        Returns
        -------
        bool
            A boolean value
        """
        return await self.media_save(media_id, collection_pk, revert=True)



import time
from uuid import uuid4
from datetime import datetime
from typing import List, Optional, Any, Dict
from .utils import generate_android_device_id, generate_uuid, generate_user_agent
from pydantic import BaseModel, FilePath, HttpUrl, ValidationError, validator


def validate_external_url(cls, v : str):
    if v is None or (v.startswith("http") and "://" in v) or isinstance(v, str):
        return v
    raise ValidationError("external_url must been URL or string")

class AuthorizationData(BaseModel):
    ds_user_id: Optional[str]
    sessionid: Optional[str]
    should_use_header_over_cookies : bool = True

class DeviceSettings(BaseModel, frozen=True):
    app_version: str = "313.0.0.0.60"
    android_version: int = 29
    android_release: str = "10.0.0"
    dpi: str = "420dpi"
    resolution: str = "1080x2104"
    manufacturer: str = "samsung"
    device: str = "xcoverpro"
    model: str = "SM-G715U"
    cpu: str = "exynos9611"
    version_code: str = "547897021"


class AccountUUIDs(BaseModel, frozen=True):
    phone_id: str = generate_uuid()
    uuid: str = generate_uuid()
    client_session_id: str = generate_uuid()
    advertising_id: str = generate_uuid()
    comment_session_id : str = generate_uuid()
    request_id: str = generate_uuid()
    tray_session_id: str = generate_uuid()
    android_device_id: str = generate_android_device_id()

class Credentials(BaseModel):
    uuids : AccountUUIDs = AccountUUIDs() 
    ig_u_rur: Optional[str] = None
    ig_www_claim: Optional[str] = None
    mid: Optional[str] = None
    shbts: Optional[str] = None
    shbid: Optional[str] = None
    region_hint: Optional[str] = None
    authorization_data: Optional[AuthorizationData] = None
    cookies: Dict[str, Any] = {}
    last_login: float = time.time()
    device_settings: DeviceSettings = DeviceSettings()
    country: str = 'US'
    country_code: int = 1
    locale: str = 'en_US'
    user_agent: str = generate_user_agent(dict(device_settings.dict(), locale=locale))
    timezone_offset: int = -21600

class Resource(BaseModel, frozen=True):
    pk: str
    video_url: Optional[HttpUrl]  # for Video and IGTV
    thumbnail_url: HttpUrl
    media_type: int


class User(BaseModel, frozen=True):
    pk: str
    username: str
    full_name: str
    is_private: bool
    profile_pic_url: HttpUrl
    profile_pic_url_hd: Optional[HttpUrl]
    is_verified: bool
    media_count: int
    follower_count: int
    following_count: int
    biography: Optional[str] = ""
    external_url: Optional[str]
    account_type: Optional[int]
    is_business: bool

    public_email: Optional[str]
    contact_phone_number: Optional[str]
    public_phone_country_code: Optional[str]
    public_phone_number: Optional[str]
    business_contact_method: Optional[str]
    business_category_name: Optional[str]
    category_name: Optional[str]
    category: Optional[str]

    address_street: Optional[str]
    city_id: Optional[str]
    city_name: Optional[str]
    latitude: Optional[float]
    longitude: Optional[float]
    zip: Optional[str]
    instagram_location_id: Optional[str]
    interop_messaging_user_fbid: Optional[str]

    _external_url = validator("external_url", allow_reuse=True)(validate_external_url)


class Account(BaseModel, frozen=True):
    pk: str
    username: str
    full_name: str
    is_private: bool
    profile_pic_url: HttpUrl
    is_verified: bool
    biography: Optional[str] = ""
    external_url: Optional[str]
    is_business: bool
    birthday: Optional[str]
    phone_number: Optional[str]
    gender: Optional[int]
    email: Optional[str]

    _external_url = validator("external_url", allow_reuse=True)(validate_external_url)


class UserShort(BaseModel, frozen=True):
    pk: str
    username: Optional[str]
    full_name: Optional[str] = ""
    profile_pic_url: Optional[HttpUrl]
    profile_pic_url_hd: Optional[HttpUrl]
    is_private: Optional[bool]
    # is_verified: bool  # not found in hashtag_medias_v1
    stories: List = []
    latest_reel_media : Optional[int]= None
    

class Usertag(BaseModel, frozen=True):
    user: UserShort
    x: float
    y: float


class Location(BaseModel, frozen=True):
    pk: Optional[int]
    name: str
    phone: Optional[str] = ""
    website: Optional[str] = ""
    category: Optional[str] = ""
    hours: Optional[dict] = {}  # opening hours
    address: Optional[str] = ""
    city: Optional[str] = ""
    zip: Optional[str] = ""
    lng: Optional[float]
    lat: Optional[float]
    external_id: Optional[int]
    external_id_source: Optional[str]
    # address_json: Optional[dict] = {}
    # profile_pic_url: Optional[HttpUrl]
    # directory: Optional[dict] = {}


class Media(BaseModel, frozen=True):
    pk: str
    id: str
    code: str
    taken_at: datetime
    media_type: int
    image_versions2: Optional[dict] = {}
    product_type: Optional[str] = ""  # igtv or feed
    thumbnail_url: Optional[HttpUrl]
    location: Optional[Location] = None
    user: UserShort
    comment_count: Optional[int] = 0
    comments_disabled: Optional[bool] = False
    commenting_disabled_for_viewer: Optional[bool] = False
    like_count: int
    play_count: Optional[int]
    has_liked: Optional[bool]
    caption_text: str
    accessibility_caption: Optional[str]
    usertags: List[Usertag]
    sponsor_tags: List[UserShort]
    video_url: Optional[HttpUrl]  # for Video and IGTV
    view_count: Optional[int] = 0  # for Video and IGTV
    video_duration: Optional[float] = 0.0  # for Video and IGTV
    title: Optional[str] = ""
    resources: List[Resource] = []
    clips_metadata: dict = {}


class MediaXma(BaseModel, frozen=True):
    # media_type: int
    video_url: Optional[HttpUrl]  # for Video and IGTV
    title: Optional[str] = ""
    preview_url: Optional[HttpUrl]
    preview_url_mime_type: Optional[str]
    header_icon_url: Optional[HttpUrl]
    header_icon_width: Optional[int]
    header_icon_height: Optional[int]
    header_title_text: Optional[str]
    preview_media_fbid: Optional[str]


class MediaOembed(BaseModel, frozen=True):
    title: str
    author_name: str
    author_url: str
    author_id: str
    media_id: str
    provider_name: str
    provider_url: HttpUrl
    type: str
    width: Optional[int] = None
    height: Optional[int] = None
    html: str
    thumbnail_url: HttpUrl
    thumbnail_width: int
    thumbnail_height: int
    can_view: bool


class Collection(BaseModel, frozen=True):
    id: str
    name: str
    type: str
    media_count: int


class Comment(BaseModel, frozen=True):
    pk: str
    text: str
    user: UserShort
    created_at_utc: datetime
    content_type: str
    status: str
    has_liked: Optional[bool]
    like_count: Optional[int]


class Hashtag(BaseModel, frozen=True):
    id: str
    name: str
    media_count: Optional[int]
    profile_pic_url: Optional[HttpUrl]


class StoryMention(BaseModel, frozen=True):
    user: UserShort
    x: Optional[float]
    y: Optional[float]
    width: Optional[float]
    height: Optional[float]


class StoryMedia(BaseModel, frozen=True):
    # Instagram does not return the feed_media object when requesting story,
    # so you will have to make an additional request to get media and this is overhead:
    # media: Media
    x: float = 0.5
    y: float = 0.4997396
    z: float = 0
    width: float = 0.8
    height: float = 0.60572916
    rotation: float = 0.0
    is_pinned: Optional[bool]
    is_hidden: Optional[bool]
    is_sticker: Optional[bool]
    is_fb_sticker: Optional[bool]
    media_pk: int
    user_id: Optional[int]
    product_type: Optional[str]
    media_code: Optional[str]


class StoryHashtag(BaseModel, frozen=True):
    hashtag: Hashtag
    x: Optional[float]
    y: Optional[float]
    width: Optional[float]
    height: Optional[float]


class StoryLocation(BaseModel, frozen=True):
    location: Location
    x: Optional[float]
    y: Optional[float]
    width: Optional[float]
    height: Optional[float]


class StoryStickerLink(BaseModel, frozen=True):
    url: HttpUrl
    link_title: Optional[str]
    link_type: Optional[str]
    display_url: Optional[str]


class StorySticker(BaseModel, frozen=True):
    id: Optional[str]
    type: Optional[str] = "gif"
    x: float
    y: float
    z: Optional[int] = 1000005
    width: float
    height: float
    rotation: Optional[float] = 0.0
    story_link: Optional[StoryStickerLink]
    extra: Optional[dict] = {}


class StoryBuild(BaseModel, frozen=True):
    mentions: List[StoryMention]
    path: FilePath
    paths: List[FilePath] = []
    stickers: List[StorySticker] = []


class StoryLink(BaseModel, frozen=True):
    webUri: HttpUrl
    x: float = 0.5126011
    y: float = 0.5168225
    z: float = 0.0
    width: float = 0.50998676
    height: float = 0.25875
    rotation: float = 0.0


class Story(BaseModel, frozen=True):
    pk: str
    id: str
    code: str
    taken_at: datetime
    media_type: int
    product_type: Optional[str] = ""
    thumbnail_url: Optional[HttpUrl]
    user: UserShort
    video_url: Optional[HttpUrl]  # for Video and IGTV
    video_duration: Optional[float] = 0.0  # for Video and IGTV
    sponsor_tags: List[UserShort]
    mentions: List[StoryMention]
    links: List[StoryLink]
    hashtags: List[StoryHashtag]
    locations: List[StoryLocation]
    stickers: List[StorySticker]
    medias: List[StoryMedia] = []


class Guide(BaseModel, frozen=True):
    id: Optional[str]
    title: Optional[str]
    description: str
    cover_media: Media
    feedback_item: Optional[dict]


class DirectMedia(BaseModel, frozen=True):
    id: str
    media_type: int
    user: Optional[UserShort]
    thumbnail_url: Optional[HttpUrl]
    video_url: Optional[HttpUrl]
    audio_url: Optional[HttpUrl]


class ReplyMessage(BaseModel, frozen=True):
    id: str
    user_id: Optional[str]
    timestamp: datetime
    item_type: Optional[str]
    is_sent_by_viewer: Optional[bool]
    is_shh_mode: Optional[bool]
    text: Optional[str]
    link: Optional[dict]
    animated_media: Optional[dict]
    media: Optional[DirectMedia]
    visual_media: Optional[dict]
    media_share: Optional[Media]
    reel_share: Optional[dict]
    story_share: Optional[dict]
    felix_share: Optional[dict]
    xma_share: Optional[MediaXma]
    clip: Optional[Media]
    placeholder: Optional[dict]


class DirectMessage(BaseModel, frozen=True):
    id: str  # e.g. 28597946203914980615241927545176064
    user_id: Optional[str]
    thread_id: Optional[int]  # e.g. 340282366841710300949128531777654287254
    timestamp: datetime
    item_type: Optional[str]
    is_sent_by_viewer: Optional[bool]
    is_shh_mode: Optional[bool]
    reactions: Optional[dict]
    text: Optional[str]
    reply: Optional[ReplyMessage]
    link: Optional[dict]
    animated_media: Optional[dict]
    media: Optional[DirectMedia]
    visual_media: Optional[dict]
    media_share: Optional[Media]
    reel_share: Optional[dict]
    story_share: Optional[dict]
    felix_share: Optional[dict]
    xma_share: Optional[MediaXma]
    clip: Optional[Media]
    placeholder: Optional[dict]


class DirectResponse(BaseModel, frozen=True):
    unseen_count: Optional[int]
    unseen_count_ts: Optional[int]
    status: Optional[str]


class DirectShortThread(BaseModel, frozen=True):
    id: str
    users: List[UserShort]
    named: bool
    thread_title: str
    pending: bool
    thread_type: str
    viewer_id: str
    is_group: bool


class DirectThread(BaseModel, frozen=True):
    pk: str  # thread_v2_id, e.g. 17898572618026348
    id: str  # thread_id, e.g. 340282366841510300949128268610842297468
    messages: List[DirectMessage]
    users: List[UserShort]
    inviter: Optional[UserShort]
    left_users: List[UserShort] = []
    admin_user_ids: list
    last_activity_at: datetime
    muted: bool
    is_pin: Optional[bool]
    named: bool
    canonical: bool
    pending: bool
    archived: bool
    thread_type: str
    thread_title: str
    folder: int
    vc_muted: bool
    is_group: bool
    mentions_muted: bool
    approval_required_for_new_members: bool
    input_mode: int
    business_thread_folder: int
    read_state: int
    is_close_friend_thread: bool
    assigned_admin_id: int
    shh_mode_enabled: bool
    last_seen_at: Optional[dict] = {}

    def is_seen(self, user_id: str):
        """Have I seen this thread?
        :param user_id: You account user_id
        """
        user_id = str(user_id)
        own_timestamp = int(self.last_seen_at[user_id]["timestamp"])
        timestamps = [
            (int(v["timestamp"]) - own_timestamp) > 0
            for k, v in self.last_seen_at.items()
            if k != user_id
        ]
        return not any(timestamps)


class Relationship(BaseModel, frozen=True):
    user_id: str
    blocking: bool
    followed_by: bool
    following: bool
    incoming_request: bool
    is_bestie: bool
    is_blocking_reel: bool
    is_muting_reel: bool
    is_private: bool
    is_restricted: bool
    muting: bool
    outgoing_request: bool


class RelationshipShort(BaseModel, frozen=True):
    user_id: str
    following: bool
    incoming_request: bool
    is_bestie: bool
    is_feed_favorite: bool
    is_private: bool
    is_restricted: bool
    outgoing_request: bool


class Highlight(BaseModel, frozen=True):
    pk: str  # 17895485401104052
    id: str  # highlight:17895485401104052
    latest_reel_media: int
    cover_media: dict
    user: UserShort
    title: str
    created_at: datetime
    is_pinned_highlight: bool
    media_count: int
    media_ids: List[int] = []
    items: List[Story] = []


class Share(BaseModel, frozen=True):
    pk: str
    type: str


class Track(BaseModel, frozen=True):
    id: str
    title: str
    subtitle: str
    display_artist: str
    audio_cluster_id: int
    artist_id: Optional[int]
    cover_artwork_uri: Optional[HttpUrl]
    cover_artwork_thumbnail_uri: Optional[HttpUrl]
    progressive_download_url: Optional[HttpUrl]
    fast_start_progressive_download_url: Optional[HttpUrl]
    reactive_audio_download_url: Optional[HttpUrl]
    highlight_start_times_in_ms: List[int]
    is_explicit: bool
    dash_manifest: str
    uri: Optional[HttpUrl]
    has_lyrics: bool
    audio_asset_id: int
    duration_in_ms: int
    dark_message: Optional[str]
    allows_saving: bool
    territory_validity_periods: dict


class Note(BaseModel, frozen=True):
    id: str
    text: str
    user_id: str
    user: UserShort
    audience: int
    created_at: datetime
    expires_at: datetime
    is_emoji_only: bool
    has_translation: bool
    note_style: int



import random
from typing import List, Optional, Tuple

from aioinstagrapi.utils import gen_user_breadcrumb, generate_uuid, generate_jazoest, dumps
from aioinstagrapi.exceptions import ClientError, ClientNotFoundError, MediaNotFound
from aioinstagrapi.extractors import extract_comment
from aioinstagrapi.models import Comment


class CommentMixin:
    """
    Helpers for managing comments on a Media
    """

    async def pre_comment(self, media_id : str, text : str):
            await self.generate_nudge(media_id)
            await self.media_check_offensive_comment(media_id, text)

    async def generate_nudge(self, media_id : str):
        data = {
            'is_bottom_sheet_open': 'true',
            'media_id': str(media_id),
            'source': 'loading',
            '_uuid': self.uuid,
            'viewed_comment_ids': '[]',
        }
        return await self.private_request('nudges/generate_nudge/', data=data)


    async def media_comments(self, media_id: str, amount: int = 20) -> List[Comment]:
        """
        Get comments on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        amount: int, optional
            Maximum number of comments to return, default is 0 - Inf

        Returns
        -------
        List[Comment]
            A list of objects of Comment
        """

        # TODO: to public or private
        def get_comments():
            if result.get("comments"):
                for comment in result.get("comments"):
                    comments.append(extract_comment(comment))

        media_id = await self.media_id(media_id)
        params = None
        comments = []
        result = await self.private_request(f"media/{media_id}/comments/", params)
        get_comments()
        while (result.get("has_more_comments") and result.get("next_max_id")) or (
            result.get("has_more_headload_comments") and result.get("next_min_id")
        ):
            try:
                if result.get("has_more_comments"):
                    params = {"max_id": result.get("next_max_id")}
                else:
                    params = {"min_id": result.get("next_min_id")}
                if not (
                    result.get("next_max_id")
                    or result.get("next_min_id")
                    or result.get("comments")
                ):
                    break
                result = await self.private_request(f"media/{media_id}/comments/", params)
                get_comments()
            except ClientNotFoundError as e:
                raise MediaNotFound(e, media_id=media_id, **self.last_json)
            except ClientError as e:
                if "Media not found" in str(e):
                    raise MediaNotFound(e, media_id=media_id, **self.last_json)
                raise e
            if amount and len(comments) >= amount:
                break
        if amount:
            comments = comments[:amount]
        return comments

    async def media_comments_chunk(
        self, media_id: str, max_amount: int, min_id: str = None
    ) -> Tuple[List[Comment], str]:
        """
        Get chunk of comments on a media and end_cursor

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        max_amount: int
            Limit number of comments to fetch, default is 100
        min_id: str, optional
            End Cursor of previous chunk that had more comments, default value is None

        Returns
        -------
        Tuple[List[Comment], str]
            A list of objects of Comment and an end_cursor
        """

        # TODO: to public or private
        def get_comments():
            if result.get("comments"):
                for comment in result.get("comments"):
                    comments.append(extract_comment(comment))

        media_id = await self.media_id(media_id)
        params = {"min_id": min_id} if min_id else None
        comments = []
        result = await self.private_request(f"media/{media_id}/comments/", params)
        get_comments()
        while result.get("has_more_headload_comments") and result.get("next_min_id"):
            try:
                params = {"min_id": result.get("next_min_id")}
                if not (result.get("next_min_id") or result.get("comments")):
                    break
                result = await self.private_request(f"media/{media_id}/comments/", params)
                get_comments()
            except ClientNotFoundError as e:
                raise MediaNotFound(e, media_id=media_id, **self.last_json)
            except ClientError as e:
                if "Media not found" in str(e):
                    raise MediaNotFound(e, media_id=media_id, **self.last_json)
                raise e
            if len(comments) >= max_amount:
                break
        return (comments, result.get("next_min_id"))

    async def media_comment(
        self, media_id: str, text: str, replied_to_comment_id: Optional[int] = None
    , gif_media_id : Optional[str] = None) -> Comment:
        """
        Post a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        text: str
            String to be posted on the media

        Returns
        -------
        Comment
            An object of Comment type
        """
        assert self.user_id, "Login required"
        media_id = await self.media_id(media_id)
        data = {
        "include_media_code":"false",
        "user_breadcrumb":gen_user_breadcrumb(len(text)),
        "inventory_source":"media_or_ad",
        "comment_creation_key":generate_jazoest(self.phone_id),
        "delivery_class":"organic",
        "idempotence_token":generate_uuid(),
        "include_e2ee_mentioned_user_list":"false",
        "radio_type":"wifi-none",
        "_uid":self.user_id,
        "_uuid":self.uuid,
        "nav_chain":"MainFeedFragment:feed_timeline:1:cold_start:1698798344.620::,InteractivityBottomSheetFragment:feed_timeline:3:button:1698798356.51::,CommentThreadFragment:comments_v2:4:button:1698798356.489::",
        "logging_info_token":"d8904f04ed084c65ac6b0f4fba074b7f",
        "comment_text":text,
        "is_carousel_bumped_post":"false",
        "container_module":"comments_v2_feed_timeline",
        "feed_position":"0"
        }
        if gif_media_id:
            data['gif_params'] = dumps({'gif_media_id': gif_media_id, 'is_sticker': False })
            #data['gif_params'] = "{\"gif_media_id\":\"WotHgjtqxZwfS\",\"is_sticker\":false}"
           

        if replied_to_comment_id:
            data["replied_to_comment_id"] = int(replied_to_comment_id)
        result = await self.private_request(
            f"media/{media_id}/comment/",
            self.with_action_data(data),
        )
        return extract_comment(result["comment"])

    async def media_check_offensive_comment(self, media_id: str, text: str) -> bool:
        """
        Checks if a comment text is offensive

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        text: str
            String to be posted on the media

        Returns
        -------
        bool
            If comment is offensive
        """
        assert self.user_id, "Login required"
        media_id = await self.media_id(media_id)
        data = {
            # _uid, comment_session_id are not in this body?
            "media_id": media_id,
            "comment_text": text,
        }
        result = await self.private_request(
            "media/comment/check_offensive_comment/",
            self.with_action_data(data),
        )
        return result["is_offensive"]

    async def comment_like(self, comment_pk: int, revert: bool = False) -> bool:
        """
        Like a comment on a media

        Parameters
        ----------
        comment_pk: int
            Unique identifier of a Comment
        revert: bool, optional
            If liked, whether or not to unlike. Default is False

        Returns
        -------
        bool
            A boolean value
        """
        assert self.user_id, "Login required"
        comment_pk = int(comment_pk)
        data = {
            "is_carousel_bumped_post": "false",
            "container_module": "feed_contextual_self_profile",
            "feed_position": str(random.randint(0, 6)),
        }
        name = "unlike" if revert else "like"
        result = await self.private_request(
            f"media/{comment_pk}/comment_{name}/", self.with_action_data(data)
        )
        return result["status"] == "ok"

    async def comment_unlike(self, comment_pk: int) -> bool:
        """
        Unlike a comment on a media

        Parameters
        ----------
        comment_pk: int
            Unique identifier of a Comment

        Returns
        -------
        bool
            A boolean value
        """
        return await self.comment_like(comment_pk, revert=True)

    async def comment_pin(self, media_id: str, comment_pk: int, revert: bool = False):
        """
        Pin a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        comment_pk: int
           Unique identifier of a Comment
        revert: bool, optional
            Unpin when True
        Returns
        -------
        bool
           A boolean value
        """
        data = self.with_action_data({"_uid": self.user_id, "_uuid": self.uuid})
        name = "unpin" if revert else "pin"

        result = await self.private_request(
            f"media/{media_id}/{name}_comment/{comment_pk}", data
        )
        return result["status"] == "ok"

    async def comment_unpin(self, media_id: str, comment_pk: int):
        """
        Unpin a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        comment_pk: int
           Unique identifier of a Comment

        Returns
        -------
        bool
           A boolean value
        """
        return await self.comment_pin(media_id, comment_pk, True)

    async def comment_bulk_delete(self, media_id: str, comment_pks: List[int]) -> bool:
        """
        Delete a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        comment_pks: List[int]
            List of unique identifier of a Comment

        Returns
        -------
        bool
            A boolean value
        """
        media_id = await self.media_id(media_id)
        data = {
            "comment_ids_to_delete": ",".join([str(pk) for pk in comment_pks]),
            "container_module": "self_comments_v2_newsfeed_you",
        }
        result = await self.private_request(
            f"media/{media_id}/comment/bulk_delete/", self.with_action_data(data)
        )
        return result["status"] == "ok"



MUTE_ALL_ITEMS = ("cancel", "15_minutes", "1_hour", "2_hour", "4_hour", "8_hour")
SETTING_VALUE_ITEMS = ("off", "following_only", "everyone")

try:
    from typing import Literal

    MUTE_ALL = Literal[MUTE_ALL_ITEMS]
    SETTING_VALUE = Literal[SETTING_VALUE_ITEMS]
except ImportError:
    # python <= 3.8
    MUTE_ALL = str
    SETTING_VALUE = str


class NotificationMixin:
    """
    Helpers for notification settings
    """

    async def notification_settings(self, content_type: str, setting_value: str) -> bool:
        data = {
            "content_type": content_type,
            "setting_value": setting_value,
            "_uid": str(self.user_id),
            "_uuid": self.uuid,
        }
        result = await self.private_request(
            "notifications/change_notification_settings/", data=data
        )
        return result.get("status") == "ok"

    async def notification_disable(self) -> bool:
        """
        Disable All Notification

        Returns
        -------
        bool
        """
        notifications = (
            self.notification_likes,
            self.notification_like_and_comment_on_photo_user_tagged,
            self.notification_user_tagged,
            self.notification_comments,
            self.notification_comment_likes,
            self.notification_first_post,
            self.notification_new_follower,
            self.notification_follow_request_accepted,
            self.notification_connection,
            self.notification_tagged_in_bio,
            self.notification_pending_direct_share,
            self.notification_direct_share_activity,
            self.notification_direct_group_requests,
            self.notification_video_call,
            self.notification_rooms,
            self.notification_live_broadcast,
            self.notification_felix_upload_result,
            self.notification_view_count,
            self.notification_fundraiser_creator,
            self.notification_fundraiser_supporter,
            self.notification_reminders,
            self.notification_announcements,
            self.notification_report_updated,
            self.notification_login,
        )
        return all((await func("off")) for func in notifications)

    async def notification_mute_all(self, setting_value: MUTE_ALL = "8_hour") -> bool:
        """
        Manage Mute All Notification Settings

        Parameters
        ----------
        setting_value: MUTE_ALL
            Value of settings, default "8_hour"

        Returns
        -------
        bool
        """
        assert (
            setting_value in MUTE_ALL_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {MUTE_ALL_ITEMS}'
        return await  self.notification_settings("mute_all", setting_value)

    async def notification_likes(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Likes Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("likes", setting_value)

    async def notification_like_and_comment_on_photo_user_tagged(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Like And Comment On Photo User Tagged Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings(
            "like_and_comment_on_photo_user_tagged", setting_value
        )

    async def notification_user_tagged(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage User Tagged NotificationSettings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("user_tagged", setting_value)

    async def notification_comments(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Comments Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("comments", setting_value)

    async def notification_comment_likes(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Comment Likes Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("comment_likes", setting_value)

    async def notification_first_post(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage First Post Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("first_post", setting_value)

    async def notification_new_follower(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage New Follower Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await  self.notification_settings("new_follower", setting_value)

    async def notification_follow_request_accepted(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Follow Request Accepted Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("follow_request_accepted", setting_value)

    async def notification_connection(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Connection Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("connection_notification", setting_value)

    async def notification_tagged_in_bio(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Tagged In Bio Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("tagged_in_bio", setting_value)

    async def notification_pending_direct_share(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Pending Direct Share Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("pending_direct_share", setting_value)

    async def notification_direct_share_activity(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Direct Share Activity Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("direct_share_activity", setting_value)

    async def notification_direct_group_requests(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Direct Group Requests Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("direct_group_requests", setting_value)

    async def notification_video_call(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Video Call Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("video_call", setting_value)

    async def notification_rooms(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Rooms Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("rooms", setting_value)

    async def notification_live_broadcast(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Live Broadcast Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("live_broadcast", setting_value)

    async def notification_felix_upload_result(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Felix Upload Result Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("felix_upload_result", setting_value)

    async def notification_view_count(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage View Count Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("view_count", setting_value)

    async def notification_fundraiser_creator(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Fundraiser Creator Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("fundraiser_creator", setting_value)

    async def notification_fundraiser_supporter(
        self, setting_value: SETTING_VALUE = "off"
    ) -> bool:
        """
        Manage Fundraiser Supporter Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("fundraiser_supporter", setting_value)

    async def notification_reminders(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Notification Reminders Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("notification_reminders", setting_value)

    async def notification_announcements(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Announcements Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("announcements", setting_value)

    async def notification_report_updated(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Report Updated Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("report_updated", setting_value)

    async def notification_login(self, setting_value: SETTING_VALUE = "off") -> bool:
        """
        Manage Login Notification Settings

        Parameters
        ----------
        setting_value: SETTING_VALUE
            Value of settings, default "off"

        Returns
        -------
        bool
        """
        assert (
            setting_value in SETTING_VALUE_ITEMS
        ), f'Unsupported setting_value="{setting_value}" {SETTING_VALUE_ITEMS}'
        return await self.notification_settings("login_notification", setting_value)



import base64
import time
import httpx
from Cryptodome.Cipher import AES, PKCS1_v1_5
from Cryptodome.PublicKey import RSA
from curlify2 import Curlify
from Cryptodome.Random import get_random_bytes


class PasswordMixin:
    async def password_encrypt(self, password):
        publickeyid, publickey = await self.password_publickeys()
        session_key = get_random_bytes(32)
        iv = get_random_bytes(12)
        timestamp = str(int(time.time()))
        decoded_publickey = base64.b64decode(publickey.encode())
        recipient_key = RSA.import_key(decoded_publickey)
        cipher_rsa = PKCS1_v1_5.new(recipient_key)
        rsa_encrypted = cipher_rsa.encrypt(session_key)
        cipher_aes = AES.new(session_key, AES.MODE_GCM, iv)
        cipher_aes.update(timestamp.encode())
        aes_encrypted, tag = cipher_aes.encrypt_and_digest(password.encode("utf8"))
        size_buffer = len(rsa_encrypted).to_bytes(2, byteorder="little")
        payload = base64.b64encode(
            b"".join(
                [
                    b"\x01",
                    publickeyid.to_bytes(1, byteorder="big"),
                    iv,
                    size_buffer,
                    rsa_encrypted,
                    tag,
                    aes_encrypted,
                ]
            )
        )
        return f"#PWD_INSTAGRAM:4:{timestamp}:{payload.decode()}"

    async def password_publickeys(self):
        headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-US',
            'Connection': 'Keep-Alive',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15',
        }
        async with httpx.AsyncClient() as client:
            resp = await client.get("https://i.instagram.com/api/v1/qe/sync/", headers=headers)
            publickeyid = int(resp.headers.get("ig-set-password-encryption-key-id"))
            publickey = resp.headers.get("ig-set-password-encryption-pub-key")
            return publickeyid, publickey
        



import logging
from urllib.parse import urlparse
from .models import Credentials
from aioinstagrapi.mixins.private import PrivateRequestMixin
from aioinstagrapi.mixins.auth.login import LoginMixin
from aioinstagrapi.mixins.user import UserMixin
from aioinstagrapi.mixins.account import AccountMixin
from aioinstagrapi.mixins.helpers import HelperMixin
from aioinstagrapi.mixins.album import DownloadAlbumMixin, UploadAlbumMixin
from aioinstagrapi.mixins.bloks import BloksMixin
from aioinstagrapi.mixins.challenge import ChallengeResolveMixin
from aioinstagrapi.mixins.clip import DownloadClipMixin, UploadClipMixin
from aioinstagrapi.mixins.collection import CollectionMixin
from aioinstagrapi.mixins.comment import CommentMixin
from aioinstagrapi.mixins.direct import DirectMixin
from aioinstagrapi.mixins.explore import ExploreMixin
from aioinstagrapi.mixins.fbsearch import FbSearchMixin
from aioinstagrapi.mixins.fundraiser import FundraiserMixin
from aioinstagrapi.mixins.hashtag import HashtagMixin
from aioinstagrapi.mixins.highlight import HighlightMixin
from aioinstagrapi.mixins.igtv import DownloadIGTVMixin, UploadIGTVMixin
from aioinstagrapi.mixins.insights import InsightsMixin
from aioinstagrapi.mixins.location import LocationMixin
from aioinstagrapi.mixins.media import MediaMixin
from aioinstagrapi.mixins.multiple_accounts import MultipleAccountsMixin
from aioinstagrapi.mixins.note import NoteMixin
from aioinstagrapi.mixins.notification import NotificationMixin
from aioinstagrapi.mixins.password import PasswordMixin
from aioinstagrapi.mixins.photo import DownloadPhotoMixin, UploadPhotoMixin
from aioinstagrapi.mixins.public import (
    ProfilePublicMixin,
    PublicRequestMixin,
    TopSearchesPublicMixin,
)
from aioinstagrapi.mixins.share import ShareMixin
from aioinstagrapi.mixins.story import StoryMixin
from aioinstagrapi.mixins.timeline import ReelsMixin
from aioinstagrapi.mixins.totp import TOTPMixin
from aioinstagrapi.mixins.track import TrackMixin
from aioinstagrapi.mixins.video import DownloadVideoMixin, UploadVideoMixin
#from requests.packages.urllib3.exceptions import InsecureRequestWarning


#requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

logging.basicConfig(level = logging.INFO)
DEFAULT_LOGGER = logging.getLogger("aioinstagrapi")
httpx_logger = logging.getLogger('httpx')
httpx_logger.propagate = False


class Client(
    PublicRequestMixin,
    ChallengeResolveMixin,
    PrivateRequestMixin,
    TopSearchesPublicMixin,
    ProfilePublicMixin,
    LoginMixin,
    ShareMixin,
    TrackMixin,
    FbSearchMixin,
    HighlightMixin,
    DownloadPhotoMixin,
    UploadPhotoMixin,
    DownloadVideoMixin,
    UploadVideoMixin,
    DownloadAlbumMixin,
    NotificationMixin,
    UploadAlbumMixin,
    DownloadIGTVMixin,
    UploadIGTVMixin,
    MediaMixin,
    UserMixin,
    InsightsMixin,
    CollectionMixin,
    AccountMixin,
    DirectMixin,
    LocationMixin,
    HashtagMixin,
    CommentMixin,
    HelperMixin,
    StoryMixin,
    PasswordMixin,
    DownloadClipMixin,
    UploadClipMixin,
    ReelsMixin,
    ExploreMixin,
    BloksMixin,
    TOTPMixin,
    MultipleAccountsMixin,
    NoteMixin,
    FundraiserMixin,
):
    proxy = None
    credentials = None

    def __init__(
        self,
        credentials: dict = {},
        proxy: str = None,
        delay_range: list = None,
        logger=DEFAULT_LOGGER,
        **kwargs,
    ):

        super().__init__(**kwargs)

        self.credentials = Credentials(**credentials)
        self.logger = logger
        self.delay_range = delay_range
        self.set_proxy(proxy)
        self.init()
    
    
    def init(self) -> bool:
        """
        Initialize credentials

        Returns
        -------
        bool
            A boolean value
        """
        self.set_uuids(self.credentials.uuids.dict())
        self.bloks_versioning_id = "525157204187ff40ed117cb8039807a0b12137bc9d50211ed3b2e1d97be928a6"
        self.app_id = "567067343352427"
        self.cookies = self.credentials.cookies
        self.authorization_data = self.credentials.authorization_data
        self.last_login = self.credentials.last_login
        self.timezone_offset = self.credentials.timezone_offset
        self.device_settings = self.credentials.device_settings
        self.user_agent = self.credentials.user_agent
        self.locale = self.credentials.locale
        self.country = self.credentials.country
        self.country_code = self.credentials.country_code
        self.mid = self.credentials.mid
        self.ig_u_rur = self.credentials.ig_u_rur
        self.shbts = self.credentials.shbts
        self.shbid = self.credentials.shbid
        self.region_hint = self.credentials.region_hint
        self.ig_www_claim = self.credentials.ig_www_claim
        return True



    def set_uuids(self, uuids: dict = None) -> bool:
        """
        Helper to set uuids

        Parameters
        ----------
        uuids: Dict, optional
            UUIDs, default is None

        Returns
        -------
        bool
            A boolean value
        """
        self.phone_id = uuids.get("phone_id")
        self.uuid = uuids.get("uuid")
        self.client_session_id = uuids.get("client_session_id")
        self.advertising_id = uuids.get("advertising_id")
        self.android_device_id = uuids.get("android_device_id")
        self.request_id = uuids.get("request_id")
        self.tray_session_id = uuids.get("tray_session_id")
        self.comment_session_id = uuids.get("comment_session_id")
        return True


import time
import asyncio
from typing import Dict, List

from aioinstagrapi.exceptions import ClientError, MediaError, UserError
from aioinstagrapi.utils import json_value

POST_TYPES = ("ALL", "CAROUSEL_V2", "IMAGE", "SHOPPING", "VIDEO")
TIME_FRAMES = (
    "ONE_WEEK",
    "ONE_MONTH",
    "THREE_MONTHS",
    "SIX_MONTHS",
    "ONE_YEAR",
    "TWO_YEARS",
)
DATA_ORDERS = (
    "REACH_COUNT",
    "LIKE_COUNT",
    "FOLLOW",
    "SHARE_COUNT",
    "BIO_LINK_CLICK",
    "COMMENT_COUNT",
    "IMPRESSION_COUNT",
    "PROFILE_VIEW",
    "VIDEO_VIEW_COUNT",
    "SAVE_COUNT",
)

try:
    from typing import Literal

    POST_TYPE = Literal[POST_TYPES]
    TIME_FRAME = Literal[TIME_FRAMES]
    DATA_ORDERING = Literal[DATA_ORDERS]
except ImportError:
    # python <= 3.8
    POST_TYPE = TIME_FRAME = DATA_ORDERING = str


class InsightsMixin:
    """
    Helper class to get insights
    """

    async def insights_media_feed_all(
        self,
        post_type: POST_TYPE = "ALL",
        time_frame: TIME_FRAME = "TWO_YEARS",
        data_ordering: DATA_ORDERING = "REACH_COUNT",
        count: int = 0,
        sleep: int = 2,
    ) -> List[Dict]:
        """
        Get insights for all medias from feed with page iteration with cursor and sleep timeout

        Parameters
        ----------
        post_type: str, optional
            Types of posts, default is "ALL"
        time_frame: str, optional
            Time frame to pull media insights, default is "TWO_YEARS"
        data_ordering: str, optional
            Ordering strategy for the data, default is "REACH_COUNT"
        count: int, optional
            Max media count for retrieving, default is 0
        sleep: int, optional
            Timeout between pages iterations, default is 2

        Returns
        -------
        List[Dict]
            List of dictionaries of response from the call
        """
        assert (
            post_type in POST_TYPES
        ), f'Unsupported post_type="{post_type}" {POST_TYPES}'
        assert (
            time_frame in TIME_FRAMES
        ), f'Unsupported time_frame="{time_frame}" {TIME_FRAMES}'
        assert (
            data_ordering in DATA_ORDERS
        ), f'Unsupported data_ordering="{data_ordering}" {DATA_ORDERS}'
        assert self.user_id, "Login required"
        medias = []
        cursor = None
        data = {
            "surface": "post_grid",
            "doc_id": 2345520318892697,
            "locale": "en_US",
            "vc_policy": "insights_policy",
            "strip_nulls": False,
            "strip_defaults": False,
        }
        query_params = {
            "IgInsightsGridMediaImage_SIZE": 480,
            "count": 200,  # TODO Try to detect max allowed value
            # "cursor": "0",
            "dataOrdering": data_ordering,
            "postType": post_type,
            "timeframe": time_frame,
            "search_base": "USER",
            "is_user": "true",
            "queryParams": {"access_token": "", "id": self.user_id},
        }
        while True:
            if cursor:
                query_params["cursor"] = cursor
            result = await self.private_request(
                "ads/graphql/",
                self.with_query_params(data, query_params),
            )
            if not json_value(
                result,
                "data",
                "shadow_instagram_user",
                "business_manager",
                default=None,
            ):
                raise UserError("Account is not business account", **self.last_json)
            stats = json_value(
                result,
                "data",
                "shadow_instagram_user",
                "business_manager",
                "top_posts_unit",
                "top_posts",
            )
            cursor = stats["page_info"]["end_cursor"]
            medias.extend(stats["edges"])
            if not stats["page_info"]["has_next_page"]:
                break
            if count and len(medias) >= count:
                break
            await asyncio.sleep(sleep)
        if count:
            medias = medias[:count]
        return medias

    """
    Helpers for getting insights for media
    """

    async def insights_account(self) -> Dict:
        """
        Get insights for account

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        assert self.user_id, "Login required"
        data = {
            "surface": "account",
            "doc_id": 2449243051851783,
            "locale": "en_US",
            "vc_policy": "insights_policy",
            "strip_nulls": False,
            "strip_defaults": False,
        }
        query_params = {
            "IgInsightsGridMediaImage_SIZE": 360,
            "activityTab": True,
            "audienceTab": True,
            "contentTab": True,
            "query_params": {"access_token": "", "id": self.user_id},
        }

        result = await self.private_request(
            "ads/graphql/",
            self.with_query_params(data, query_params),
        )
        res = json_value(result, "data", "shadow_instagram_user", "business_manager")
        if not res:
            raise UserError("Account is not business account", **self.last_json)
        return res

    async def insights_media(self, media_pk: int) -> Dict:
        """
        Get insights data for media

        Parameters
        ----------
        media_pk: int
            PK for the album you want to download

        Returns
        -------
        Dict
            A dictionary with insights data
        """
        assert self.user_id, "Login required"
        media_pk = self.media_pk(media_pk)
        data = {
            "surface": "post",
            "doc_id": 3221905377882880,
            "locale": "en_US",
            "vc_policy": "insights_policy",
            "strip_nulls": False,
            "strip_defaults": False,
        }
        query_params = {
            "query_params": {"access_token": "", "id": media_pk},
        }
        try:
            result = await self.private_request(
                "ads/graphql/",
                self.with_query_params(data, query_params),
            )
            return result["data"]["instagram_post_by_igid"]
        except ClientError as e:
            raise MediaError(e.message, media_pk=media_pk, **self.last_json)



import base64
import time
import re
import json
from pathlib import Path
from typing import Dict
from uuid import uuid4
from typing import Union
from aioinstagrapi.models import AuthorizationData
from pydantic import ValidationError
from aioinstagrapi.mixins.auth.prelog import PreLoginFlowMixin
from aioinstagrapi.mixins.auth.postlog import PostLoginFlowMixin
from aioinstagrapi.exceptions import *
from aioinstagrapi.utils import gen_token, dumps, generate_jazoest
from aioinstagrapi.models import AuthorizationData, Credentials

class LoginMixin(PreLoginFlowMixin, PostLoginFlowMixin):
    username = None
    password = None
    authorization_data = None
    last_login = None
    relogin_attempt = 0
    device_settings = {}
    client_session_id = ""
    tray_session_id = ""
    advertising_id = ""
    android_device_id = ""
    comment_session_id = ""
    request_id = ""
    phone_id = ""
    app_id = "567067343352427"
    uuid = ""
    mid = "0"
    country = "US"
    country_code = 1  
    locale = "en_US"
    timezone_offset: int = -18000  
    ig_u_rur = ""  
    ig_www_claim = ""  


    def __init__(self, username : str = None, password : str = None):
        self.username = username
        self.password = password

    
    @property
    def authorization(self) -> str:
        """Build authorization header
        Example: Bearer IGT:2:eaW9u.....aWQiOiI0NzM5=
        """
        if self.authorization_data:
            b64part = base64.b64encode(dumps(self.authorization_data.dict()).encode()).decode()
            return f"Bearer IGT:2:{b64part}"
        return ""

    @property
    def sessionid(self) -> str:
        sessionid = self.cookies.get("sessionid")
        if not sessionid and self.authorization_data:
            sessionid = self.authorization_data.dict().get("sessionid")
        return sessionid

    @property
    def token(self) -> str:
        """CSRF token
        e.g. vUJGjpst6szjI38mZ6Pb1dROsWVerZelGSYGe0W1tuugpSUefVjRLj2Pom2SWNoA
        """
        if not getattr(self, "_token", None):
            self._token = self.cookies.get("csrftoken", gen_token(64))
        return self._token

    @property
    def rank_token(self) -> str:
        return f"{self.user_id}_{self.uuid}"

    @property
    def user_id(self) -> int:
        user_id = self.cookies.get("ds_user_id")
        if not user_id and self.authorization_data:
            user_id = self.authorization_data.dict().get("ds_user_id")
        if user_id:
            return int(user_id)
        return None

    @property
    def device(self) -> dict:
        return {
            key: val
            for key, val in self.device_settings.dict().items()
            if key in ["manufacturer", "model", "android_version", "android_release"]
        }

    async def login(
        self,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
        relogin: bool = False,
        verification_code: str = "",
    ) -> bool:
        """
        Login

        Parameters
        ----------
        username: str
            Instagram Username
        password: str
            Instagram Password
        relogin: bool
            Whether or not to re login, default False
        verification_code: str
            2FA verification code

        Returns
        -------
        bool
            A boolean value
        """

        if not self.username or not self.password:
            if username is None or password is None:
                raise BadCredentials("Both username and password must be provided.")

            self.username = username
            self.password = password

        if relogin:
            self.authorization_data = None
            self.cookies.clear()
            if self.relogin_attempt > 1:
                raise ReloginAttemptExceeded()
            self.relogin_attempt += 1
        # if self.user_id and self.last_login:
        #     if time.time() - self.last_login < 60 * 60 * 24:
        #        return True  # already login
        if self.user_id and not relogin:
            return True  # already login
        try:
            await self.pre_login_flow()
        except (PleaseWaitFewMinutes, ClientThrottledError):
            self.logger.warning("Ignore 429: Continue login")
            # The instagram application ignores this error
            # and continues to log in (repeat this behavior)
        enc_password = await self.password_encrypt(self.password)
        data = {
            "jazoest": generate_jazoest(self.phone_id),
            "country_codes": '[{"country_code":"%d","source":["default"]}]'
            % int(self.country_code),
            "phone_id": self.phone_id,
            "enc_password": enc_password,
            "username": username,
            "adid": self.advertising_id,
            "guid": self.uuid,
            "device_id": self.android_device_id,
            "google_tokens": "[]",
            "login_attempt_count": "0",
        }
        try:
            logged = await self.private_request("accounts/login/", data, login=True)
            self.authorization_data = AuthorizationData(**self.parse_authorization(
                self.last_response.headers.get("ig-set-authorization")
            ))
            self.credentials.authorization_data = self.authorization_data
        except TwoFactorRequired as e:
            if not verification_code.strip():
                raise TwoFactorRequired(
                    f"{e} (you did not provide verification_code for login method)"
                )
            two_factor_identifier = self.last_json.get("two_factor_info", {}).get(
                "two_factor_identifier"
            )
            data = {
                "verification_code": verification_code,
                "phone_id": self.phone_id,
                "_csrftoken": self.token,
                "two_factor_identifier": two_factor_identifier,
                "username": username,
                "trust_this_device": "0",
                "guid": self.uuid,
                "device_id": self.android_device_id,
                "waterfall_id": str(uuid4()),
                "verification_method": "3",
            }
            logged = await self.private_request(
                "accounts/two_factor_login/", data, login=True
            )
            self.authorization_data = AuthorizationData(**self.parse_authorization(
                self.last_response.headers.get("ig-set-authorization")
            ))
            self.credentials.authorization_data = self.authorization_data
        if logged:
            await self.login_flow()
            self.last_login = time.time()
            return True
        return False

    async def login_by_sessionid(self, sessionid: str) -> bool:
        """
        Login using session id

        Parameters
        ----------
        sessionid: str
            Session ID

        Returns
        -------
        bool
            A boolean value
        """
        assert isinstance(sessionid, str) and len(sessionid) > 30, "Invalid sessionid"
        user_id = re.search(r"^\d+", sessionid).group()
        self.credentials.cookies = {"sessionid": sessionid, 'ds_user_id': user_id}
        self.credentials.authorization_data = AuthorizationData(**{
            "ds_user_id": user_id,
            "sessionid": sessionid,
            "should_use_header_over_cookies": True,
        })
        self.init()
        try:
            user = await self.user_info_v1(int(user_id))
        except :
            raise
        self.username = user.username
        return True

    
    def inject_sessionid_to_public(self) -> bool:
        """
        Inject sessionid from private session to public session

        Returns
        -------
        bool
            A boolean value
        """
        if self.sessionid:
            self.public_cookies['sessionid']= self.sessionid
            return True
        return False

    
    async def one_tap_app_login(self, user_id: str, nonce: str) -> bool:
        """One tap login emulation

        Parameters
        ----------
        user_id: str
            User ID
        nonce: str
            Login nonce (from Instagram, e.g. in /logout/)

        Returns
        -------
        bool
            A boolean value
        """
        user_id = int(user_id)
        data = {
            "phone_id": self.phone_id,
            "user_id": user_id,
            "adid": self.advertising_id,
            "guid": self.uuid,
            "device_id": self.uuid,
            "login_nonce": nonce,
            "_csrftoken": self.token,
        }
        return await self.private_request("accounts/one_tap_app_login/", data)

    async def relogin(self) -> bool:
        """
        Relogin helper

        Returns
        -------
        bool
            A boolean value
        """
        return await self.login(self.username, self.password, relogin=True)

    def get_settings(self) -> Dict:
        """
        Get current session settings

        Returns
        -------
        Dict
            Current session settings as a Dict
        """
        data = self.credentials.dict()
        data['cookies'] = self.cookies
        return data

    
    def set_settings(self, settings: Dict) -> bool:
        """
        Set session settings

        Returns
        -------
        Bool
        """
        self.credentials = Credentials(**settings)
        self.init()
        return True

    def load_settings(self, path: Path) -> Dict:
        """
        Load session settings

        Parameters
        ----------
        path: Path
            Path to storage file

        Returns
        -------
        Dict
            Current session settings as a Dict
        """
        with open(path, "r") as fp:
            self.set_settings(json.load(fp))
            return self.credentials
        return None

    def dump_settings(self, path: Path) -> bool:
        """
        Serialize and save session settings

        Parameters
        ----------
        path: Path
            Path to storage file

        Returns
        -------
        Bool
        """
        with open(path, "w") as fp:
            json.dump(self.get_settings(), fp, indent=4)
        return True


import random
import time
import asyncio
import httpx
import aiofiles
from pathlib import Path
from typing import Dict, List
from urllib.parse import urlparse
from uuid import uuid4

import requests

from aioinstagrapi import config
from aioinstagrapi.exceptions import (
    VideoConfigureError,
    VideoConfigureStoryError,
    VideoNotDownload,
    VideoNotUpload,
)
from aioinstagrapi.extractors import extract_direct_message, extract_media_v1
from aioinstagrapi.models import (
    DirectMessage,
    Location,
    Media,
    Story,
    StoryHashtag,
    StoryLink,
    StoryLocation,
    StoryMedia,
    StoryMention,
    StorySticker,
    Usertag,
)
from aioinstagrapi.utils import date_time_original, dumps, generate_uuid, generate_mutation_token


class DownloadVideoMixin:
    """
    Helpers for downloading video
    """

    async def video_download(self, media_pk: int, folder: Path = "") -> Path:
        """
        Download video using media pk

        Parameters
        ----------
        media_pk: int
            Unique Media ID
        folder: Path, optional
            Directory in which you want to download the video, default is "" and will download the files to working dir.

        Returns
        -------
        Path
            Path for the file downloaded
        """
        media = await self.media_info(media_pk)
        assert media.media_type == 2, "Must been video"
        filename = "{username}_{media_pk}".format(
            username=media.user.username, media_pk=media_pk
        )
        return await self.video_download_by_url(media.video_url, filename, folder)

    async def video_download_by_url(
        self, url: str, filename: str = "", folder: Path = ""
    ) -> Path:
        """
        Download video using URL

        Parameters
        ----------
        url: str
            URL for a media
        filename: str, optional
            Filename for the media
        folder: Path, optional
            Directory in which you want to download the video, default is "" and will download the files to working
                directory

        Returns
        -------
        Path
            Path for the file downloaded
        """
        fname = urlparse(url).path.rsplit("/", 1)[1]
        filename = "%s.%s" % (filename, fname.rsplit(".", 1)[1]) if filename else fname
        path = Path(folder) / filename
        async with httpx.AsyncClient() as client:
            async with client.stream("GET", url, timeout=self.request_timeout) as response:
                response.raise_for_status()
                
                try:
                    content_length = int(response.headers.get("Content-Length"))
                except TypeError:
                    print("""
                        The program detected a mis-formatted link, and hence can't download it.
                        This problem occurs when the URL is passed into
                        'video_download_by_url()' or the 'clip_download_by_url()'.
                        The raw URL needs to be re-formatted into one that is recognizable by the methods.
                        Use this code: url=self.cl.media_info(self.cl.media_pk_from_url('insert the url here')).video_url
                        You can remove the 'self' from the code above if needed.
                        """)
                    raise Exception("The program detected a mis-formatted link.")
                    
                file_length = 0
                async for chunk in response.aiter_bytes():
                    file_length += len(chunk)
                    
                if content_length != file_length:
                    raise VideoNotDownload(
                        f'Broken file "{path}" (Content-length={content_length}, but file length={file_length})'
                    )
                    
                async with aiofiles.open(path, "wb") as f:
                    async for chunk in response.aiter_bytes():
                        f.write(chunk)
                        
        return path.resolve()

    async def video_download_by_url_origin(self, url: str) -> bytes:
        """
        Download video using URL

        Parameters
        ----------
        url: str
            URL for a media

        Returns
        -------
        bytes
            Bytes for the file downloaded
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            content_length = int(response.headers.get("Content-Length"))
            file_length = len(response.content)
            if content_length != file_length:
                raise VideoNotDownload(
                    f'Broken file from url "{url}" (Content-length={content_length}, but file length={file_length})'
                )
            return response.content

              


class UploadVideoMixin:
    """
    Helpers for downloading video
    """

    async def video_rupload(
        self,
        path: Path,
        thumbnail: Path = None,
        to_album: bool = False,
        to_story: bool = False,
        to_direct: bool = False,
    ) -> tuple:
        """
        Upload video to Instagram

        Parameters
        ----------
        path: Path
            Path to the media
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        to_album: bool, optional
        to_story: bool, optional
        to_direct: bool, optional

        Returns
        -------
        tuple
            (Upload ID for the media, width, height)
        """
        assert isinstance(path, Path), f"Path must been Path, now {path} ({type(path)})"
        upload_id = str(int(time.time() * 1000))
        width, height, duration, thumbnail = await analyze_video(path, thumbnail)
        waterfall_id = str(uuid4())
        # upload_name example: '1576102477530_0_7823256191'
        upload_name = "{upload_id}_0_{rand}".format(
            upload_id=upload_id, rand=random.randint(1000000000, 9999999999)
        )
        rupload_params = {
            "retry_context": '{"num_step_auto_retry":0,"num_reupload":0,"num_step_manual_retry":0}',
            "media_type": "2",
            "xsharing_user_ids": dumps([self.user_id]),
            "upload_id": upload_id,
            "upload_media_duration_ms": str(int(duration * 1000)),
            "upload_media_width": str(width),
            "upload_media_height": str(height),  # "1138" for Mi5s
        }
        if to_direct:
            rupload_params["direct_v2"] = "1"
            # "hflip": "false",
            # "rotate":"3",
        if to_album:
            rupload_params["is_sidecar"] = "1"
        if to_story:
            rupload_params = {
                "extract_cover_frame": "1",
                "content_tags": "has-overlay",
                "for_album": "1",
                **rupload_params,
            }
        headers = {
            "Accept-Encoding": "gzip, deflate",
            "X-Instagram-Rupload-Params": dumps(rupload_params),
            "X_FB_VIDEO_WATERFALL_ID": waterfall_id,
            # "X_FB_VIDEO_WATERFALL_ID": "88732215909430_55CF262450C9_Mixed_0",  # ALBUM
            # "X_FB_VIDEO_WATERFALL_ID": "1594919079102",  # VIDEO
        }
        if to_album:
            headers = {"Segment-Start-Offset": "0", "Segment-Type": "3", **headers}
        response = await self.private_request(
            domain="https://{domain}/rupload_igvideo/{name}".format(
                domain=config.API_DOMAIN, name=upload_name
            ),
            headers=headers,
        )
        if self.last_response.status_code != 200:
            raise VideoNotUpload(str(response), response=self.last_response, **self.last_json)
        async with aiofiles.open(path, "rb") as fp:
            video_data = fp.read()
            video_len = str(len(video_data))
        headers = {
            "Offset": "0",
            "X-Entity-Name": upload_name,
            "X-Entity-Length": video_len,
            "Content-Type": "application/octet-stream",
            "Content-Length": video_len,
            "X-Entity-Type": "video/mp4",
            **headers,
        }
        response = await self.private_request(
            domain="https://{domain}/rupload_igvideo/{name}".format(
                domain=config.API_DOMAIN, name=upload_name
            ),
            data=video_data,
            headers=headers,
        )
        if self.last_response.status_code != 200:
            raise VideoNotUpload(str(response), response=self.last_response, **self.last_json)
        return upload_id, width, height, duration, Path(thumbnail)

    async def video_upload(
        self,
        path: Path,
        caption: str,
        thumbnail: Path = None,
        usertags: List[Usertag] = [],
        location: Location = None,
        extra_data: Dict[str, str] = {},
    ) -> Media:
        """
        Upload video and configure to feed

        Parameters
        ----------
        path: Path
            Path to the media
        caption: str
            Media caption
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        usertags: List[Usertag], optional
            List of users to be tagged on this upload, default is empty list.
        location: Location, optional
            Location tag for this upload, default is None
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Media
            An object of Media class
        """
        path = Path(path)
        if thumbnail is not None:
            thumbnail = Path(thumbnail)
        upload_id, width, height, duration, thumbnail = await self.video_rupload(
            path, thumbnail, to_story=False
        )
        for attempt in range(50):
            self.logger.debug(f"Attempt #{attempt} to configure Video: {path}")
            await asyncio.sleep(3)
            try:
                configured = await self.video_configure(
                    upload_id,
                    width,
                    height,
                    duration,
                    thumbnail,
                    caption,
                    usertags,
                    location,
                    extra_data=extra_data,
                )
            except Exception as e:
                if "Transcode not finished yet" in str(e):
                    """
                    Response 202 status:
                    {"message": "Transcode not finished yet.", "status": "fail"}
                    """
                    await asyncio.sleep(10)
                    continue
                raise e
            else:
                if configured:
                    media = configured.get("media")
                    await self.expose()
                    return extract_media_v1(media)
        raise VideoConfigureError(response=self.last_response, **self.last_json)

    async def video_configure(
        self,
        upload_id: str,
        width: int,
        height: int,
        duration: int,
        thumbnail: Path,
        caption: str,
        usertags: List[Usertag] = [],
        location: Location = None,
        extra_data: Dict[str, str] = {},
    ) -> Dict:
        """
        Post Configure Video (send caption, thumbnail and more to Instagram)

        Parameters
        ----------
        upload_id: str
            Unique upload_id
        width: int
            Width of the video in pixels
        height: int
            Height of the video in pixels
        duration: int
            Duration of the video in seconds
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        caption: str
            Media caption
        usertags: List[Usertag], optional
            List of users to be tagged on this upload, default is empty list.
        location: Location, optional
            Location tag for this upload, default is None
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        await self.photo_rupload(Path(thumbnail), upload_id)
        usertags = [
            {"user_id": tag.user.pk, "position": [tag.x, tag.y]} for tag in usertags
        ]
        data = {
            "multi_sharing": "1",
            "creation_logger_session_id": self.client_session_id,
            "upload_id": upload_id,
            "source_type": "4",
            "location": self.location_build(location),
            "poster_frame_index": 0,
            "length": duration,
            "audio_muted": False,
            "usertags": dumps({"in": usertags}),
            "filter_type": "0",
            "date_time_original": date_time_original(time.localtime()),
            "timezone_offset": str(self.timezone_offset),
            "clips": [{"length": duration, "source_type": "4"}],
            "extra": {"source_width": width, "source_height": height},
            "device": self.device,
            "caption": caption,
            **extra_data,
        }
        return await self.private_request(
            "media/configure/?video=1", self.with_default_data(data)
        )

    async def video_upload_to_story(
        self,
        path: Path,
        caption: str = "",
        thumbnail: Path = None,
        mentions: List[StoryMention] = [],
        locations: List[StoryLocation] = [],
        links: List[StoryLink] = [],
        hashtags: List[StoryHashtag] = [],
        stickers: List[StorySticker] = [],
        medias: List[StoryMedia] = [],
        extra_data: Dict[str, str] = {},
    ) -> Story:
        """
        Upload video as a story and configure it

        Parameters
        ----------
        path: Path
            Path to the media
        caption: str
            Story caption
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        mentions: List[StoryMention], optional
            List of mentions to be tagged on this upload, default is empty list.
        locations: List[StoryLocation], optional
            List of locations to be tagged on this upload, default is empty list.
        links: List[StoryLink]
            URLs for Swipe Up
        hashtags: List[StoryHashtag], optional
            List of hashtags to be tagged on this upload, default is empty list.
        stickers: List[StorySticker], optional
            List of stickers to be tagged on this upload, default is empty list.
        medias: List[StoryMedia], optional
            List of medias to be tagged on this upload, default is empty list.
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Story
            An object of Media class
        """
        path = Path(path)
        if thumbnail is not None:
            thumbnail = Path(thumbnail)
        upload_id, width, height, duration, thumbnail = await self.video_rupload(
            path, thumbnail, to_story=True
        )
        for attempt in range(50):
            self.logger.debug(f"Attempt #{attempt} to configure Video: {path}")
            await asyncio.sleep(3)
            try:
                configured = await self.video_configure_to_story(
                    upload_id,
                    width,
                    height,
                    duration,
                    thumbnail,
                    caption,
                    mentions,
                    locations,
                    links,
                    hashtags,
                    stickers,
                    medias,
                    extra_data=extra_data,
                )
            except Exception as e:
                if "Transcode not finished yet" in str(e):
                    """
                    Response 202 status:
                    {"message": "Transcode not finished yet.", "status": "fail"}
                    """
                    await asyncio.sleep(10)
                    continue
                raise e
            if configured:
                media = configured.get("media")
                await self.expose()
                return Story(
                    links=links,
                    mentions=mentions,
                    hashtags=hashtags,
                    locations=locations,
                    stickers=stickers,
                    medias=medias,
                    **extract_media_v1(media).dict(),
                )
        raise VideoConfigureStoryError(response=self.last_response, **self.last_json)

    async def video_configure_to_story(
        self,
        upload_id: str,
        width: int,
        height: int,
        duration: int,
        thumbnail: Path,
        caption: str,
        mentions: List[StoryMention] = [],
        locations: List[StoryLocation] = [],
        links: List[StoryLink] = [],
        hashtags: List[StoryHashtag] = [],
        stickers: List[StorySticker] = [],
        medias: List[StoryMedia] = [],
        thread_ids: List[int] = [],
        extra_data: Dict[str, str] = {},
    ) -> Dict:
        """
        Story Configure for Photo

        Parameters
        ----------
        upload_id: str
            Unique upload_id
        width: int
            Width of the video in pixels
        height: int
            Height of the video in pixels
        duration: int
            Duration of the video in seconds
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        caption: str
            Media caption
        mentions: List[StoryMention], optional
            List of mentions to be tagged on this upload, default is empty list.
        locations: List[StoryLocation], optional
            List of locations to be tagged on this upload, default is empty list.
        links: List[StoryLink]
            URLs for Swipe Up
        hashtags: List[StoryHashtag], optional
            List of hashtags to be tagged on this upload, default is empty list.
        stickers: List[StorySticker], optional
            List of stickers to be tagged on this upload, default is empty list.
        medias: List[StoryMedia], optional
            List of medias to be tagged on this upload, default is empty list.
        thread_ids: List[int], optional
            List of Direct Message Thread ID (to send a story to a thread)
        extra_data: Dict[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        timestamp = int(time.time())
        mentions = mentions.copy()
        locations = locations.copy()
        links = links.copy()
        hashtags = hashtags.copy()
        stickers = stickers.copy()
        medias = medias.copy()
        thread_ids = thread_ids.copy()
        story_sticker_ids = []
        data = {
            # USE extra_data TO EXTEND THE SETTINGS OF THE LOADED STORY,
            #   USE FOR EXAMPLE THE PROPERTIES SPECIFIED IN THE COMMENT:
            # ---------------------------------
            # When send to DIRECT:
            # "allow_multi_configures": "1",
            # "client_context":"6823316152962778207",
            #      ^----- token = random.randint(6800011111111111111, 6800099999999999999) from direct.py
            # "is_shh_mode":"0",
            # "mutation_token":"6824688191453546273",
            # "nav_chain":"1qT:feed_timeline:1,1qT:feed_timeline:7,ReelViewerFragment:reel_feed_timeline:21,5HT:attribution_quick_camera_fragment:22,4ji:reel_composer_preview:23,8wg:direct_story_audience_picker:24,4ij:reel_composer_camera:25,ReelViewerFragment:reel_feed_timeline:26",
            # "recipient_users":"[]",
            # "send_attribution":"direct_story_audience_picker",
            # "thread_ids":"[\"340282366841710300949128149448121770626\"]",  <-- send story to direct
            # "view_mode": "replayable",
            # ---------------------------------
            # Optional (markup for caption field) when tagging:
            # "story_captions":"[{\"text\":\"@user1+\\n\\n@user2+\",\"position_data\":{\"x\":0.5,\"y\":0.5,\"height\":272.0,\"width\":670.0,\"rotation\":0.0},\"scale\":1.0,\"font_size\":24.0,\"format_type\":\"classic_v2\",\"effects\":[\"disabled\"],\"colors\":[\"#ffffff\"],\"alignment\":\"center\",\"animation\":\"\"}]",
            # ---------------------------------
            # SEGMENT MODE (when file is too big):
            # "allow_multi_configures": "1",
            # "segmented_video_group_id": str(uuid4()),
            # "multi_upload_session_id": str(uuid4()),
            # "segmented_video_count": "4",  # "4"  # SEGMENT MODE
            # "segmented_video_index": "0",  # 0,1,2,3  # SEGMENT MODE
            # "is_multi_upload": "1",  # SEGMENT MODE
            # "is_segmented_video": "1",  # SEGMENT MODE
            # ---------------------------------
            # COMMON properties:
            "_uid": str(self.user_id),
            "supported_capabilities_new": dumps(config.SUPPORTED_CAPABILITIES),
            "has_original_sound": "1",
            "filter_type": "0",
            "camera_session_id": self.client_session_id,
            "camera_entry_point": str(random.randint(35, 164)),
            "composition_id": generate_uuid(),
            # "camera_make": self.device_settings.get("manufacturer", "Xiaomi"),
            # "camera_model": self.device_settings.get("model", "MI+5s"),
            "timezone_offset": str(self.timezone_offset),
            "client_timestamp": str(timestamp),
            "client_shared_at": str(timestamp - 7),  # 7 seconds ago
            # "imported_taken_at": str(timestamp - 5 * 24 * 3600),  # 5 days ago
            "date_time_original": date_time_original(time.localtime()),
            # "date_time_digitalized": date_time_original(time.localtime()),
            # "story_sticker_ids": "",
            # "media_folder": "Camera",
            "configure_mode": "1",
            # "configure_mode": "2", <- when direct
            "source_type": "3",  # "3"
            "video_result": "",
            "creation_surface": "camera",
            # "software": config.SOFTWARE.format(**self.device_settings),
            # "caption": caption,
            "capture_type": "normal",
            # "rich_text_format_types": '["classic_v2"]',  # default, typewriter
            "upload_id": upload_id,
            # "scene_capture_type": "standard",
            # "scene_type": "",
            "original_media_type": "video",
            "camera_position": "back",
            # Facebook Sharing Part:
            # "xpost_surface": "auto_xpost",
            # "share_to_fb_destination_type": "USER",
            # "share_to_fb_destination_id":"832928543",
            # "share_to_facebook":"1",
            # "fb_access_token":"EAABwzLixnjYBACVgqBfLyDuPWs6RN2sTZC........cnNkjHCH2",
            # "attempt_id": str(uuid4()),
            "device": self.device,
            "length": duration,
            "clips": [
                {"length": duration, "source_type": "3", "camera_position": "back"}
            ],
            # "edits": {
            #     "filter_type": 0,
            #     "filter_strength": 1.0,
            #     "crop_original_size": [width, height],
            #     # "crop_center": [0, 0],
            #     # "crop_zoom": 1
            # },
            "media_transformation_info": dumps(
                {
                    "width": str(width),
                    "height": str(height),
                    "x_transform": "0",
                    "y_transform": "0",
                    "zoom": "1.0",
                    "rotation": "0.0",
                    "background_coverage": "0.0",
                }
            ),
            "extra": {"source_width": width, "source_height": height},
            "audio_muted": False,
            "poster_frame_index": 0,
            # "app_attribution_android_namespace": "",
        }
        data.update(extra_data)
        tap_models = []
        static_models = []
        if mentions:
            reel_mentions = []
            text_metadata = []
            for mention in mentions:
                reel_mentions.append(
                    {
                        "x": mention.x,
                        "y": mention.y,
                        "z": 0,
                        "width": mention.width,
                        "height": mention.height,
                        "rotation": 0.0,
                        "type": "mention",
                        "user_id": str(mention.user.pk),
                        "is_sticker": False,
                        "display_type": "mention_username",
                        "tap_state": 0,
                        "tap_state_str_id": "mention_text",
                    }
                )
                text_metadata.append(
                    {
                        "font_size": 24.0,
                        "scale": 1.0,
                        "width": 366.0,
                        "height": 102.0,
                        "x": mention.x,
                        "y": mention.y,
                        "rotation": 0.0,
                    }
                )
            data["text_metadata"] = dumps(text_metadata)
            # data["reel_mentions"] = dumps(reel_mentions)
            tap_models.extend(reel_mentions)
        if hashtags:
            story_sticker_ids.append("hashtag_sticker")
            for mention in hashtags:
                item = {
                    "x": mention.x,
                    "y": mention.y,
                    "z": 0,
                    "width": mention.width,
                    "height": mention.height,
                    "rotation": 0.0,
                    "type": "hashtag",
                    "tag_name": mention.hashtag.name,
                    "is_sticker": True,
                    "tap_state": 0,
                    "tap_state_str_id": "hashtag_sticker_gradient",
                }
                tap_models.append(item)
        if locations:
            story_sticker_ids.append("location_sticker")
            for mention in locations:
                mention.location = self.location_complete(mention.location)
                item = {
                    "x": mention.x,
                    "y": mention.y,
                    "z": 0,
                    "width": mention.width,
                    "height": mention.height,
                    "rotation": 0.0,
                    "type": "location",
                    "location_id": str(mention.location.pk),
                    "is_sticker": True,
                    "tap_state": 0,
                    "tap_state_str_id": "location_sticker_vibrant",
                }
                tap_models.append(item)
        if links:
            # instagram allow one link now
            link = links[0]
            self.private_request(
                "media/validate_reel_url/",
                {
                    "url": str(link.webUri),
                    "_uid": str(self.user_id),
                    "_uuid": str(self.uuid),
                },
            )
            stickers.append(
                StorySticker(
                    type="story_link",
                    x=link.x,
                    y=link.y,
                    z=link.z,
                    width=link.width,
                    height=link.height,
                    rotation=link.rotation,
                    extra=dict(
                        link_type="web",
                        url=str(link.webUri),
                        tap_state_str_id="link_sticker_default",
                    ),
                )
            )
            story_sticker_ids.append("link_sticker_default")
        if stickers:
            for sticker in stickers:
                sticker_extra = sticker.extra or {}
                if sticker.id:
                    sticker_extra["str_id"] = sticker.id
                    story_sticker_ids.append(sticker.id)
                tap_models.append(
                    {
                        "x": round(sticker.x, 7),
                        "y": round(sticker.y, 7),
                        "z": sticker.z,
                        "width": round(sticker.width, 7),
                        "height": round(sticker.height, 7),
                        "rotation": sticker.rotation,
                        "type": sticker.type,
                        "is_sticker": True,
                        "selected_index": 0,
                        "tap_state": 0,
                        **sticker_extra,
                    }
                )
                if sticker.type == "gif":
                    data["has_animated_sticker"] = "1"
        if medias:
            for feed_media in medias:
                assert feed_media.media_pk, "Required StoryMedia.media_pk"
                # if not feed_media.user_id:
                #     user = self.media_user(feed_media.media_pk)
                #     feed_media.user_id = user.pk
                item = {
                    "x": feed_media.x,
                    "y": feed_media.y,
                    "z": feed_media.z,
                    "width": feed_media.width,
                    "height": feed_media.height,
                    "rotation": feed_media.rotation,
                    "type": "feed_media",
                    "media_id": str(feed_media.media_pk),
                    "media_owner_id": str(feed_media.user_id or ""),
                    "product_type": "feed",
                    "is_sticker": True,
                    "tap_state": 0,
                    "tap_state_str_id": "feed_post_sticker_square",
                }
                tap_models.append(item)
            data["reshared_media_id"] = str(feed_media.media_pk)
        if thread_ids:
            # Send to direct thread
            token = generate_mutation_token()
            data.update(
                {
                    "configure_mode": "2",
                    "allow_multi_configures": "1",
                    "client_context": token,
                    "is_shh_mode": "0",
                    "mutation_token": token,
                    "nav_chain": (
                        "1qT:feed_timeline:1,1qT:feed_timeline:7,ReelViewerFragment:reel_feed_timeline:21,"
                        "5HT:attribution_quick_camera_fragment:22,4ji:reel_composer_preview:23,"
                        "8wg:direct_story_audience_picker:24,4ij:reel_composer_camera:25,"
                        "ReelViewerFragment:reel_feed_timeline:26"
                    ),
                    "recipient_users": "[]",
                    "send_attribution": "direct_story_audience_picker",
                    "thread_ids": dumps([str(tid) for tid in thread_ids]),
                    "view_mode": "replayable",
                }
            )
        if tap_models:
            data["tap_models"] = dumps(tap_models)
        if static_models:
            data["static_models"] = dumps(static_models)
        if story_sticker_ids:
            data["story_sticker_ids"] = story_sticker_ids[0]
        return await self.private_request(
            "media/configure_to_story/?video=1", self.with_default_data(data)
        )

    async def video_upload_to_direct(
        self,
        path: Path,
        caption: str = "",
        thumbnail: Path = None,
        mentions: List[StoryMention] = [],
        medias: List[StoryMedia] = [],
        thread_ids: List[int] = [],
        extra_data: Dict[str, str] = {},
    ) -> DirectMessage:
        """
        Upload video to direct thread as a story and configure it

        Parameters
        ----------
        path: Path
            Path to the media
        caption: str
            Story caption
        thumbnail: str
            Path to thumbnail for video. When None, then thumbnail is generate automatically
        mentions: List[StoryMention], optional
            List of mentions to be tagged on this upload, default is empty list.
        thread_ids: List[int], optional
            List of Direct Message Thread ID (to send a story to a thread)
        extra_data: List[str, str], optional
            Dict of extra data, if you need to add your params, like {"share_to_facebook": 1}.

        Returns
        -------
        DirectMessage
            An object of DirectMessage class
        """
        path = Path(path)
        if thumbnail is not None:
            thumbnail = Path(thumbnail)
        upload_id, width, height, duration, thumbnail = await self.video_rupload(
            path, thumbnail, to_story=True
        )
        for attempt in range(50):
            self.logger.debug(f"Attempt #{attempt} to configure Video: {path}")
            await asyncio.sleep(3)
            try:
                configured = await self.video_configure_to_story(
                    upload_id,
                    width,
                    height,
                    duration,
                    thumbnail,
                    caption,
                    mentions=mentions,
                    medias=medias,
                    thread_ids=thread_ids,
                    extra_data=extra_data,
                )
            except Exception as e:
                if "Transcode not finished yet" in str(e):
                    """
                    Response 202 status:
                    {"message": "Transcode not finished yet.", "status": "fail"}
                    """
                    await asyncio.sleep(10)
                    continue
                raise e
            if configured and thread_ids:
                return extract_direct_message(configured.get("message_metadata", [])[0])
        raise VideoConfigureStoryError(response=self.last_response, **self.last_json)


async def analyze_video(path: Path, thumbnail: Path = None) -> tuple:
    """
    Story Configure for Photo

    Parameters
    ----------
    path: Path
        Path to the media
    thumbnail: str
        Path to thumbnail for video. When None, then thumbnail is generate automatically

    Returns
    -------
    Tuple
        (width, height, duration, thumbnail)
    """

    try:
        import moviepy.editor as mp
    except ImportError:
        raise Exception("Please install moviepy>=1.0.3 and retry")

    print(f'Analyzing video file "{path}"')
    video = mp.VideoFileClip(str(path))
    width, height = video.size
    if not thumbnail:
        thumbnail = f"{path}.jpg"
        print(f'Generating thumbnail "{thumbnail}"...')
        video.save_frame(thumbnail, t=(video.duration / 2))
    # duration = round(video.duration + 0.001, 3)
    try:
        video.close()
    except AttributeError:
        pass
    return width, height, video.duration, thumbnail



import json
import base64
from typing import Dict

class HelperMixin:

    def parse_authorization(self, authorization) -> dict:
        """Parse authorization header"""
        try:
            b64part = authorization.rsplit(":", 1)[-1]
            if not b64part:
                return {}
            return json.loads(base64.b64decode(b64part))
        except Exception as e:
            self.logger.exception(e)
        return {}
    
    def with_extra_data(self, data: Dict) -> Dict:
        """
        Helper to get extra data

        Returns
        -------
        Dict
            A dictionary of default data
        """
        return self.with_default_data(
            {
                "phone_id": self.phone_id,
                "_uid": str(self.user_id),
                "guid": self.uuid,
                **data,
            }
        )

    def with_default_data(self, data: Dict) -> Dict:
        """
        Helper to get default data

        Returns
        -------
        Dict
            A dictionary of default data
        """
        return {
            "_uuid": self.uuid,
            # "_uid": str(self.user_id),
            # "_csrftoken": self.token,
            "device_id": self.android_device_id,
            **data,
        }

    def with_action_data(self, data: Dict) -> Dict:
        """
        Helper to get action data

        Returns
        -------
        Dict
            A dictionary of action data
        """
        return dict(self.with_default_data({"radio_type": "wifi-none"}), **data)

    async def expose(self) -> Dict:
        """
        Helper to expose

        Returns
        -------
        Dict
            A dictionary of response from the call
        """
        data = {"id": self.uuid, "experiment": "ig_android_profile_contextual_feed"}
        return await self.private_request("qe/expose/", self.with_default_data(data))


import json
import logging
import random
import time
import httpx
import traceback
import asyncio
import ssl
from urllib.parse import urlparse
from httpx import Timeout
from json.decoder import JSONDecodeError
from aioconsole import ainput
from aioinstagrapi.utils import *
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_random



from aioinstagrapi import config
from aioinstagrapi.exceptions import (
    BadPassword,
    ScrapingBlock,
    InternalServerError,
    GenericComments,
    ClientSuspended,
    ClientNoGifPermissions,
    ChallengeRequired,
    ClientBadRequestError,
    ClientConnectionError,
    ClientError,
    ClientForbiddenError,
    PrivateAccount,
    CommentsDisabledError,
    UserNotFound,
    ProxyAddressIsBlocked,
    InvalidTargetUser,
    InvalidMediaId,
    MediaUnavailable,
    ClientJSONDecodeError,
    ClientNotFoundError,
    ClientRequestTimeout,
    ClientThrottledError,
    FeedbackRequired,
    LoginRequired,
    PleaseWaitFewMinutes,
    RateLimitError,
    SentryBlock,
    TwoFactorRequired,
    UnknownError,
    VideoTooLongException,
)
from aioinstagrapi.utils import dumps, generate_signature

class Response:
    status_code : int
    text : str
    json : dict


async def manual_input_code(self, username: str, choice=None):
    """
    Manual security code helper

    Parameters
    ----------
    username: str
        User name of a Instagram account
    choice: optional
        Whether sms or email

    Returns
    -------
    str
        Code
    """
    code = None
    while True:
        code = (await ainput(f"Enter code (6 digits) for {username} ({choice}): ")).strip()
        if code and code.isdigit():
            break
    return code # is not int, because it can start from 0


async def manual_change_password(self, username: str):
    pwd = None
    while not pwd:
        pwd = await ainput(f"Enter password for {username}: ")
    return pwd.strip()


class PrivateRequestMixin:
    """
    Helpers for private request
    """

    private_requests_count = 0
    handle_exception = None
    challenge_code_handler = manual_input_code
    change_password_handler = manual_change_password
    private_request_logger = logging.getLogger("private_request")
    request_timeout = 5
    domain = config.API_DOMAIN
    last_response = None
    last_json = {}
    proxies = {}
    cookies = {}


    def __init__(self, *args, **kwargs):
        
        self.request_timeout = kwargs.pop("request_timeout", self.request_timeout)
        super().__init__(*args, **kwargs)


    async def small_delay(self):
        await asyncio.sleep(random.uniform(0.75, 3.75))

    async def very_small_delay(self):
        await asyncio.sleep(random.uniform(0.175, 0.875))
    
    async def random_delay(self):
        delay_function = random.choice([self.small_delay, self.very_small_delay])
        await delay_function()

    def set_proxy(self, dsn : str = None):
        if dsn:
            assert isinstance(
                dsn, str
            ), f'Proxy must been string (URL), but now "{dsn}" ({type(dsn)})'
            self.proxy = dsn
            proxy_href = "{scheme}{href}".format(
                scheme="http://" if not urlparse(self.proxy).scheme else "",
                href=self.proxy,
            )
            self.proxies = {
                "http://": proxy_href,
                "https://": proxy_href,
            }
            #input(self.proxies)
            return True
        self.proxies = {}
        return False

    @property
    def base_headers_old(self):
        locale = self.locale.replace("-", "_")
        accept_language = ["en-US"]
        if locale:
            lang = locale.replace("_", "-")
            if lang not in accept_language:
                accept_language.insert(0, lang)
        headers = {
            "X-IG-App-Locale": locale,
            "X-IG-Device-Locale": locale,
            "X-IG-Mapped-Locale": locale,
            'X-Fb-Connection-Type': 'WIFI',
            "X-Pigeon-Session-Id": generate_uuid("UFS-", "-1"),
            "X-Pigeon-Rawclienttime": str(round(time.time(), 3)),
            "X-IG-Bandwidth-Speed-KBPS": str(
                random.randint(2500000, 3000000) / 1000
            ),  # "-1.000"
            "X-IG-Bandwidth-TotalBytes-B": str(
                random.randint(5000000, 90000000)
            ),  # "0"
            "X-IG-Bandwidth-TotalTime-MS": str(random.randint(2000, 9000)),  # "0"
            "X-IG-App-Startup-Country": self.country.upper(),
            "X-Bloks-Version-Id": self.bloks_versioning_id,
            "X-IG-WWW-Claim": "0",
            "X-Bloks-Is-Layout-RTL": "false",
            "X-Bloks-Is-Panorama-Enabled": "true",
            "X-IG-Device-ID": self.uuid,
            "X-IG-Family-Device-ID": self.phone_id,
            "X-IG-Android-ID": self.android_device_id,
            "X-IG-Timezone-Offset": str(self.timezone_offset),
            "X-IG-Connection-Type": "WIFI",
            "X-IG-Capabilities": "3brTvx0=",  
            "X-IG-App-ID": self.app_id,
            "Priority": "u=3",
            "User-Agent": self.user_agent,
            "Accept-Language": ", ".join(accept_language),
            "X-MID": self.mid or "0",  
            "Accept-Encoding": "gzip, deflate", 
            "Host": self.domain,
            "X-FB-HTTP-Engine": "Liger",
            "Connection": "keep-alive",
            # "Pragma": "no-cache",
            # "Cache-Control": "no-cache",
            "X-FB-Client-IP": "True",
            "X-FB-Server-Cluster": "True",
            "IG-INTENDED-USER-ID": '0',
            "X-IG-Nav-Chain": "9MV:self_profile:2,ProfileMediaTabFragment:self_profile:3,9Xf:self_following:4",
            "X-IG-SALT-IDS": str(random.randint(1061162222, 1061262222)),
        }
        if self.user_id:
            next_year = time.time() + 31536000  # + 1 year in seconds
            headers.update(
                {
                    
                    "IG-U-DS-USER-ID": str(self.user_id),
                    "IG-INTENDED-USER-ID": str(self.user_id ),
                    # Direct:
                    "IG-U-IG-DIRECT-REGION-HINT": (
                        f"LLA,{self.user_id},{next_year}:"
                        "01f7bae7d8b131877d8e0ae1493252280d72f6d0d554447cb1dc9049b6b2c507c08605b7"
                    ),
                    "IG-U-SHBID": (
                        f"12695,{self.user_id},{next_year}:"
                        "01f778d9c9f7546cf3722578fbf9b85143cd6e5132723e5c93f40f55ca0459c8ef8a0d9f"
                    ),
                    "IG-U-SHBTS": (
                        f"{int(time.time())},{self.user_id},{next_year}:"
                        "01f7ace11925d0388080078d0282b75b8059844855da27e23c90a362270fddfb3fae7e28"
                    ),
                    "IG-U-RUR": (
                        f"RVA,{self.user_id},{next_year}:"
                        "01f7f627f9ae4ce2874b2e04463efdb184340968b1b006fa88cb4cc69a942a04201e544c"
                    ),
                }
            )
        if self.authorization_data:    
            headers.update({"Authorization": self.authorization})
        if self.ig_u_rur:
            headers.update({"IG-U-RUR": self.ig_u_rur})
        if self.ig_www_claim:
            headers.update({"X-IG-WWW-Claim": self.ig_www_claim})
        return headers


    @property
    def base_headers(self):
        locale = self.locale.replace("-", "_")
        accept_language = ["en-US"]
        if locale:
            lang = locale.replace("_", "-")
            if lang not in accept_language:
                accept_language.insert(0, lang)
        headers = {
            'Host': 'i.instagram.com',
            'X-Tigon-Is-Retry': 'False',
            'X-Ig-App-Locale': locale,
            'X-Ig-Device-Locale': locale,
            "X-MID": self.mid or "0",  
            'X-Ig-Mapped-Locale': locale,
            'X-Pigeon-Session-Id': generate_uuid("UFS-", "-1"),
            'X-Pigeon-Rawclienttime': str(round(time.time(), 3)),
            "X-IG-Bandwidth-Speed-KBPS": str(
                random.randint(2500000, 3000000) / 1000
            ),  # "-1.000"
            "X-IG-Bandwidth-TotalBytes-B": str(
                random.randint(5000000, 90000000)
            ),  # "0"
            "X-IG-Bandwidth-TotalTime-MS": str(random.randint(2000, 9000)),
            'X-Bloks-Version-Id': self.bloks_versioning_id,
            'X-Ig-Www-Claim': '0',
            'X-Bloks-Is-Prism-Enabled': 'false',
            'X-Bloks-Is-Layout-Rtl': 'false',
            "X-IG-Device-ID": self.uuid,
            "X-IG-Family-Device-ID": self.phone_id,
            "X-IG-Android-ID": self.android_device_id,
            "X-IG-Timezone-Offset": str(self.timezone_offset),
            'X-Fb-Connection-Type': 'WIFI',
            'X-Ig-Connection-Type': 'WIFI',
            'X-Ig-Capabilities': '3brTv10=',
            'X-Ig-App-Id': self.app_id,
            'Priority': 'u=3',
            'User-Agent': 'Instagram 313.0.0.0.60 Android (29/10; 420dpi; 1080x2104; samsung; SM-G715U; xcoverpro; exynos9611; en_US; 547897021)',
            'Accept-Language': 'en-US',
            'Ig-Intended-User-Id': '0',
            #'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
            # 'Content-Length': '134',
            # 'Accept-Encoding': 'gzip, deflate, br',
            'X-Fb-Http-Engine': 'Liger',
            'X-Fb-Client-Ip': 'True',
            'X-Fb-Server-Cluster': 'True',
        }
        if self.user_id:
            next_year = int(time.time() + 31535100) # + 1 year in seconds
            old = {
                "Ig-U-Ig-Direct-Region-Hint": (
                        f"RVA,{self.user_id},{next_year}:"
                        "01f70acfd56749fd883ea00d5e4063a506d77c0f6e7c170ef1f7c04212dd60fa992b27f7"
                    ),
                    "Ig-U-Shbid": (
                        f"12695,{self.user_id},{next_year}:"
                        "01f77b9eb29aa76cda03bebc81353efcd8e338a8abdebbacb6743f30fe02b2220d2438c4"
                    ),
                    "Ig-U-Shbts": (
                        f"{int(time.time())},{self.user_id},{next_year}:"
                        "01f777dfb5dc99c772be1bdcca77f292a860849f8d32f6f34ea0a2e93b6ad25c9233e25e"
                    ),
                    "Ig-U-Rur": (
                        f"EAG,{self.user_id},{next_year}:"
                        "01f73db7bda21deee42ce3b60099a5c6898c006a5fe613acb1ae9763ed17263a96be9a9f"
                    ),
            }
            headers.update(
                {
                    #"X-Ig-Nav-Chain": "9MV:self_profile:2,ProfileMediaTabFragment:self_profile:3,9Xf:self_following:4",
                    "Ig-U-Ds-User-Id": str(self.user_id),
                    "Ig-Intended-User-Id": str(self.user_id ),
                    "Ig-U-Shbid": (
                        f"12695,{self.user_id},{next_year}:"
                        "01f77b9eb29aa76cda03bebc81353efcd8e338a8abdebbacb6743f30fe02b2220d2438c4"
                    ),
                    "Ig-U-Shbts": (
                        f"{int(time.time())},{self.user_id},{next_year}:"
                        "01f777dfb5dc99c772be1bdcca77f292a860849f8d32f6f34ea0a2e93b6ad25c9233e25e"
                    ),
                    
                    # Direct:
                    
                }
            )
        if self.authorization_data:    
            headers.update({"Authorization": self.authorization})
        if self.ig_u_rur:
            headers.update({"Ig-U-Rur": self.ig_u_rur})
        if self.ig_www_claim:
            headers.update({"X-Ig-Www-Claim": self.ig_www_claim})
        if self.shbid:
            headers.update({"Ig-U-Shbid": self.shbid})
        if self.shbts:
            headers.update({"Ig-U-Shbts": self.shbts})
        if self.region_hint:
            headers.update({"Ig-U-Ig-Direct-Region-Hint": self.region_hint})
        return headers

  
    
    @staticmethod
    def with_query_params(data, params):
        return dict(data, **{"query_params": json.dumps(params, separators=(",", ":"))})

    def set_response_cookies(self, response : httpx.Response):
        #MAKE SETTERS
        mid = response.headers.get("ig-set-x-mid")
        if mid:
            self.mid = mid  
            self.credentials.mid = mid
        rur = response.headers.get("ig-set-ig-u-rur")
        if rur:
            self.ig_u_rur = rur
            self.credentials.ig_u_rur = rur
        claim = response.headers.get("x-ig-set-www-claim")  
        if claim:
            self.ig_www_claim = claim
            self.credentials.ig_www_claim = self.ig_www_claim
        shbts = response.headers.get("ig-set-ig-u-shbts")
        if shbts:
            self.credentials.shbts = shbts
            self.shbts = shbts
        shbid = response.headers.get("ig-set-ig-u-shbid")    
        if shbid:
            self.credentials.shbid = shbid
            self.shbid = shbid  
        region_hint = response.headers.get("ig-set-ig-u-ig-direct-region-hint")  
        if region_hint:
            self.credentials.region_hint = region_hint
            self.region_hint = region_hint
        authorization = response.headers.get('ig-set-authorization')
        if authorization:
            ...
            #self.authorization_data = self.parse_authorization(authorization)
    @retry(retry=retry_if_exception_type((httpx.TimeoutException, ssl.SSLSyscallError,httpx.ConnectError)), stop=stop_after_attempt(10), wait=wait_random(1,3))
    async def _send_private_request(
        self,
        endpoint='',
        data=None,
        params=None,
        login=False,
        with_signature=True,
        headers=None,
        extra_sig=None,
        domain: str = None,
    ):
        self.last_response = None
        self.last_json = last_json = {}
        req_headers = (self.base_headers)
        if headers:
            req_headers.update(headers)
        if not login:
            await asyncio.sleep(self.request_timeout)
        # if self.user_id and login:
        #     raise Exception(f"User already logged ({self.user_id})")
        try:
            if not endpoint.startswith("/"):
                endpoint = f"/v1/{endpoint}"

            if endpoint == "/challenge/":  # wow so hard, is it safe tho?
                endpoint = "/v1/challenge/"

            api_url = domain or f"https://{domain or config.API_DOMAIN}/api{endpoint}"
            self.logger.info(api_url)
            if data:  # POST
                # Client.direct_answer raw dict
                # data = json.dumps(data)
                req_headers["Content-Type"] = "application/x-www-form-urlencoded; charset=UTF-8"
                async with httpx.AsyncClient(
                    proxies=self.proxies, 
                    headers=req_headers, 
                    cookies=self.cookies
                ) as client:
                
                    if with_signature:
                        # Client.direct_answer doesn't need a signature
                        data = generate_signature(dumps(data))
                        if extra_sig:
                            data += "&".join(extra_sig)
                    tries = 0
                    while tries < 50:
                        try:
                            response = await client.post(
                                url=api_url, data=data, params=params, timeout=15
                            )
                            break
                        except Exception as e:
                            print(traceback.format_exc())
            else:  # GET
                async with httpx.AsyncClient(
                    proxies=self.proxies, 
                    headers=req_headers, 
                    cookies=self.cookies
                ) as client:
                    tries = 0
                    while tries < 50:
                        try:
                            response = await client.get(
                                url=api_url, params=params, timeout=15
                            )
                            break
                        except Exception as e:
                            print(traceback.format_exc())
            self.logger.debug(
                "private_request %s: %s (%s)",
                response.status_code,
                response.url,
                response.text,
            )
            self.set_response_cookies(response)
            self.request_log(response)
            self.last_response = response
            response.raise_for_status()
            self.last_json = last_json = response.json()
            self.logger.debug("last_json %s", last_json)
        except JSONDecodeError as e:
            self.logger.error(
                "Status %s: JSONDecodeError in private_request (user_id=%s, endpoint=%s) >>> %s",
                response.status_code,
                self.user_id,
                endpoint,
                response.text,
            )
            raise ClientJSONDecodeError(
                "JSONDecodeError {0!s} while opening {1!s}".format(e, response.url),
                response=response,
            )
        except (httpx.HTTPError, httpx.HTTPStatusError) as e:
            try:
                self.last_json = last_json = response.json()
            except:
                pass
            message = last_json.get("message", "")
            text = str(last_json)
            if "don't have permission to post a GIF" in text:
                raise ClientNoGifPermissions(e, response=e.response, **last_json)
            if "Please wait a few minutes" in message:
                raise PleaseWaitFewMinutes(e, response=e.response, **last_json)
            if "https://www.instagram.com/accounts/suspended/" in text:
                raise ClientSuspended(e, response=e.response, **last_json)
            if e.response.status_code >= 500:
                raise InternalServerError(e, response=e.response, **last_json)
            if e.response.status_code == 403:
                if message == "login_required":
                    raise LoginRequired(response=e.response, **last_json)
                if len(e.response.text) < 512:
                    last_json["message"] = e.response.text
                raise ClientForbiddenError(e, response=e.response, **last_json)
            elif e.response.status_code == 400:
                error_type = last_json.get("error_type")
                if message == "challenge_required":
                    raise ChallengeRequired(**last_json)
                if 'scraping' in text and "wait a few minutes" in text:
                    raise ScrapingBlock(e, response=e.response, **last_json)
                if "generic" in text and 'feedback_required' in text: 
                    raise GenericComments(e, response=e.response, **last_json)
                elif (
                    "Media is unavailable" in message
                    or "Media not found or unavailable" in message
                    or "comment_media_does_not_exist" in text
                ):
                    raise MediaUnavailable(e, response=e.response, **last_json)
                elif 'comment_comments_disabled' in text:
                    raise CommentsDisabledError(e, response=e.response, **last_json)
                elif message == "feedback_required":
                    raise FeedbackRequired(
                        **dict(
                            last_json,
                            message="%s: %s"
                            % (message, last_json.get("feedback_message")),
                        )
                    )
                elif error_type == "sentry_block":
                    raise SentryBlock(**last_json)
                elif error_type == "rate_limit_error":
                    raise RateLimitError(**last_json)
                elif error_type == "bad_password":
                    msg = last_json.get("message", "").strip()
                    if msg:
                        if not msg.endswith("."):
                            msg = "%s." % msg
                        msg = "%s " % msg
                    last_json["message"] = (
                        "%sIf you are sure that the password is correct, then change your IP address, "
                        "because it is added to the blacklist of the Instagram Server"
                    ) % msg
                    raise BadPassword(**last_json)
                elif error_type == "two_factor_required":
                    if not last_json["message"]:
                        last_json["message"] = "Two-factor authentication required"
                    raise TwoFactorRequired(**last_json)
                elif "VideoTooLongException" in message:
                    raise VideoTooLongException(e, response=e.response, **last_json)
                elif "Not authorized to view user" in message:
                    raise PrivateAccount(e, response=e.response, **last_json)
                elif "Invalid target user" in message:
                    raise InvalidTargetUser(e, response=e.response, **last_json)
                elif "Invalid media_id" in message:
                    raise InvalidMediaId(e, response=e.response, **last_json)
                elif "has been deleted" in message:
                    # Sorry, this photo has been deleted.
                    raise MediaUnavailable(e, response=e.response, **last_json)
                elif "unable to fetch followers" in message:
                    # returned when user not found
                    raise UserNotFound(e, response=e.response, **last_json)
                elif "The username you entered" in message:
                    # The username you entered doesn't appear to belong to an account.
                    # Please check your username and try again.
                    last_json["message"] = (
                        "Instagram has blocked your IP address, "
                        "use a quality proxy provider (not free, not shared)"
                    )
                    raise ProxyAddressIsBlocked(**last_json)
                elif error_type or message:
                    raise UnknownError(**last_json)
                # TODO: Handle last_json with {'message': 'counter get error', 'status': 'fail'}
                self.logger.exception(e)
                self.logger.warning(
                    "Status 400: %s",
                    message or "Empty response message. Maybe enabled Two-factor auth?",
                )
                raise ClientBadRequestError(e, response=e.response, **last_json)
            elif e.response.status_code == 429:
                self.logger.warning("Status 429: Too many requests")
                raise ClientThrottledError(e, response=e.response, **last_json)
            elif e.response.status_code == 404:
                self.logger.warning("Status 404: Endpoint %s does not exist", endpoint)
                raise ClientNotFoundError(e, response=e.response, **last_json)
            elif e.response.status_code == 408:
                self.logger.warning("Status 408: Request Timeout")
                raise ClientRequestTimeout(e, response=e.response, **last_json)
            raise ClientError(e, response=e.response, **last_json)
        except httpx.ConnectError as e:
            raise ClientConnectionError("{e.__class__.__name__} {e}".format(e=e))
        if last_json.get("status") == "fail":
            raise ClientError(response=response, **last_json)
        elif "error_title" in last_json:
            """Example: {
            'error_title': 'bad image input extra:{}', <-------------
            'media': {
                'device_timestamp': '1588184737203',
                'upload_id': '1588184737203'
            },
            'message': 'media_needs_reupload', <-------------
            'status': 'ok' <-------------
            }"""
            raise ClientError(response=response, **last_json)
        return last_json

    def request_log(self, response : httpx.Response):
        self.private_request_logger.info(
            "%s [%s] %s %s (%s)",
            self.username,
            response.status_code,
            response.request.method,
            response.url,
            "{app_version}, {manufacturer} {model}".format(
                app_version=self.device_settings.app_version,
                manufacturer=self.device_settings.manufacturer,
                model=self.device_settings.model,
            ),
        )

    async def private_request(
        self,
        endpoint='',
        data=None,
        params=None,
        login=False,
        with_signature=True,
        headers=None,
        extra_sig=None,
        domain: str = None,
    ):
        kwargs = dict(
            data=data,
            params=params,
            login=login,
            with_signature=with_signature,
            headers=headers,
            extra_sig=extra_sig,
            domain=domain,
        )
        try:
            await self.random_delay()
            self.private_requests_count += 1
            await self._send_private_request(endpoint, **kwargs)
        except ClientRequestTimeout:
            self.logger.info(
                "Wait 5 seconds and try one more time (ClientRequestTimeout)"
            )
            await asyncio.sleep(5)
            return await self._send_private_request(endpoint, **kwargs)
        # except BadPassword as e:
        #     raise e
        except Exception as e:
            if self.handle_exception:
                self.handle_exception(self, e)
            elif isinstance(e, ChallengeRequired):
                await self.challenge_resolve(self.last_json)
            else:
                raise e
            if login and self.user_id:
                # After challenge resolve return last_json
                return self.last_json
            return await self._send_private_request(endpoint, **kwargs)
        return self.last_json



import datetime
import enum
import json
import random
import string
import time
import urllib
import hashlib
import base64
import uuid
import hmac
from .config import USER_AGENT_BASE


class InstagramIdCodec:
    ENCODING_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"

    @staticmethod
    def encode(num, alphabet=ENCODING_CHARS):
        """Covert a numeric value to a shortcode."""
        num = int(num)
        if num == 0:
            return alphabet[0]
        arr = []
        base = len(alphabet)
        while num:
            rem = num % base
            num //= base
            arr.append(alphabet[rem])
        arr.reverse()
        return "".join(arr)

    @staticmethod
    def decode(shortcode, alphabet=ENCODING_CHARS):
        """Covert a shortcode to a numeric value."""
        base = len(alphabet)
        strlen = len(shortcode)
        num = 0
        idx = 0
        for char in shortcode:
            power = strlen - (idx + 1)
            num += alphabet.index(char) * (base**power)
            idx += 1
        return num


class InstagrapiJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, enum.Enum):
            return obj.value
        elif isinstance(obj, datetime.time):
            return obj.strftime("%H:%M")
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            return int(obj.strftime("%s"))
        elif isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def generate_signature(data):
    """Generate signature of POST data for Private API

    Returns
    -------
    str
        e.g. "signed_body=SIGNATURE.test"
    """
    return "signed_body=SIGNATURE.{data}".format(data=urllib.parse.quote_plus(data))


def json_value(data, *args, default=None):
    cur = data
    for a in args:
        try:
            if isinstance(a, int):
                cur = cur[a]
            else:
                cur = cur.get(a)
        except (IndexError, KeyError, TypeError, AttributeError):
            return default
    return cur


def gen_token(size=10, symbols=False):
    """Gen CSRF or something else token"""
    chars = string.ascii_letters + string.digits
    if symbols:
        chars += string.punctuation
    return "".join(random.choice(chars) for _ in range(size))


def gen_password(size=10):
    """Gen password"""
    return gen_token(size)


def dumps(data):
    """Json dumps format as required Instagram"""
    return InstagrapiJSONEncoder(separators=(",", ":")).encode(data)


def generate_jazoest(symbols: str) -> str:
    amount = sum(ord(s) for s in symbols)
    return f"2{amount}"


def date_time_original(localtime):
    # return time.strftime("%Y:%m:%d+%H:%M:%S", localtime)
    return time.strftime("%Y%m%dT%H%M%S.000Z", localtime)


def random_delay(delay_range: list):
    """Trigger sleep of a random floating number in range min_sleep to max_sleep"""
    return time.sleep(random.uniform(delay_range[0], delay_range[1]))


def generate_android_device_id():
        """
        Helper to generate Android Device ID
        """
        return "android-%s" % hashlib.sha256(str(time.time()).encode()).hexdigest()[:16]

def generate_uuid(prefix: str = "", suffix: str = ""):
        """
        Helper to generate uuids
        """
        return f"{prefix}{uuid.uuid4()}{suffix}"

def generate_mutation_token() -> str:
        """Token used when DM sending and upload media"""
        return str(random.randint(6800011111111111111, 6800099999999999999))

def generate_user_agent(data):
    return USER_AGENT_BASE.format(**data)


def gen_user_breadcrumb(size: int) -> str:
        """
        Helper to generate user breadcrumbs

        Parameters
        ----------
        size: int
            Integer value

        Returns
        -------
        Str
            A string
        """
        key = "iN4$aGr0m"
        dt = int(time.time() * 1000)
        time_elapsed = random.randint(500, 1500) + size * random.randint(500, 1500)
        text_change_event_count = max(1, size / random.randint(3, 5))
        data = "{size!s} {elapsed!s} {count!s} {dt!s}".format(
            **{
                "size": size,
                "elapsed": time_elapsed,
                "count": text_change_event_count,
                "dt": dt,
            }
        )
        return "{!s}\n{!s}\n".format(
            base64.b64encode(
                hmac.new(
                    key.encode("ascii"), data.encode("ascii"), digestmod=hashlib.sha256
                ).digest()
            ),
            base64.b64encode(data.encode("ascii")),
        )